"""Implement cyipopt's Interior Point Optimizer."""

from dataclasses import dataclass
from typing import Any, Literal

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import Bounds as ScipyBounds

from optimagic import mark
from optimagic.config import IS_CYIPOPT_INSTALLED
from optimagic.exceptions import NotInstalledError
from optimagic.optimization.algo_options import (
    CONVERGENCE_FTOL_REL,
    STOPPING_MAXITER,
)
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalBounds,
    InternalOptimizationProblem,
)
from optimagic.optimizers.scipy_optimizers import process_scipy_result
from optimagic.typing import (
    AggregationLevel,
    GtOneFloat,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    YesNoBool,
)

if IS_CYIPOPT_INSTALLED:
    import cyipopt


@mark.minimizer(
    name="ipopt",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_CYIPOPT_INSTALLED,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=True,
    disable_history=False,
)
@dataclass(frozen=True)
class Ipopt(Algorithm):
    # convergence criteria
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    dual_inf_tol: PositiveFloat = 1.0
    constr_viol_tol: PositiveFloat = 0.0001
    compl_inf_tol: PositiveFloat = 0.0001
    s_max: float = 100
    mu_target: NonNegativeFloat = 0.0
    # stopping criteria
    stopping_maxiter: PositiveInt = STOPPING_MAXITER
    stopping_max_wall_time_seconds: PositiveFloat = 1e20
    stopping_max_cpu_time: PositiveFloat = 1e20
    # acceptable criteria
    acceptable_iter: NonNegativeInt = 15
    acceptable_tol: PositiveFloat = 1e-6
    acceptable_dual_inf_tol: PositiveFloat = 1e-10
    acceptable_constr_viol_tol: PositiveFloat = 0.01
    acceptable_compl_inf_tol: PositiveFloat = 0.01
    acceptable_obj_change_tol: PositiveFloat = 1e20
    diverging_iterates_tol: PositiveFloat = 1e20
    nlp_lower_bound_inf: float = -1e19
    nlp_upper_bound_inf: float = 1e19
    fixed_variable_treatment: Literal[
        "make_parameter",
        "make_parameter_nodual",
        "relax_bounds",
        "make_constraint",
    ] = "make_parameter"
    dependency_detector: Literal["none", "mumps", "wsmp", "ma28"] | None = None
    dependency_detection_with_rhs: YesNoBool = False
    # bounds
    kappa_d: NonNegativeFloat = 1e-5
    bound_relax_factor: NonNegativeFloat = 1e-8
    honor_original_bounds: YesNoBool = False
    # derivatives
    check_derivatives_for_naninf: YesNoBool = False
    # not sure if we should support the following:
    jac_c_constant: YesNoBool = False
    jac_d_constant: YesNoBool = False
    hessian_constant: YesNoBool = False
    # scaling
    nlp_scaling_method: (
        Literal[
            "none",
            "user-scaling",
            "gradient-based",
            "equilibration-based",
        ]
        | None
    ) = "gradient-based"
    obj_scaling_factor: float = 1
    nlp_scaling_max_gradient: PositiveFloat = 100
    nlp_scaling_obj_target_gradient: NonNegativeFloat = 0.0
    nlp_scaling_constr_target_gradient: NonNegativeFloat = 0.0
    nlp_scaling_min_value: NonNegativeFloat = 1e-8
    # initialization
    bound_push: PositiveFloat = 0.01
    # TODO: refine type to fix the range (0,0.5]
    bound_frac: PositiveFloat = 0.01
    slack_bound_push: PositiveFloat = 0.01
    # TODO: refine type to fix the range (0,0.5]
    slack_bound_frac: PositiveFloat = 0.01
    constr_mult_init_max: NonNegativeFloat = 1000
    bound_mult_init_val: PositiveFloat = 1
    bound_mult_init_method: Literal[
        "constant",
        "mu-based",
    ] = "constant"
    least_square_init_primal: YesNoBool = False
    least_square_init_duals: YesNoBool = False
    # warm start
    warm_start_init_point: YesNoBool = False
    warm_start_same_structure: YesNoBool = False
    warm_start_bound_push: PositiveFloat = 0.001
    warm_start_bound_frac: PositiveFloat = 0.001
    warm_start_slack_bound_push: PositiveFloat = 0.001
    # TODO: refine type to fix the range (0,0.5])
    warm_start_slack_bound_frac: PositiveFloat = 0.001
    warm_start_mult_bound_push: PositiveFloat = 0.001
    warm_start_mult_init_max: float = 1e6
    warm_start_entire_iterate: YesNoBool = False
    warm_start_target_mu: float = 0.0
    # miscellaneous
    option_file_name: str = ""
    replace_bounds: YesNoBool = False
    skip_finalize_solution_call: YesNoBool = False
    timing_statistics: YesNoBool = False
    # barrier parameter update
    mu_max_fact: PositiveFloat = 1000
    mu_max: PositiveFloat = 100_000
    mu_min: PositiveFloat = 1e-11
    adaptive_mu_globalization: Literal[
        "obj-constr-filter",
        "kkt-error",
        "never-monotone-mode",
    ] = "obj-constr-filter"
    adaptive_mu_kkterror_red_iters: NonNegativeInt = 4
    # TODO: refine type to fix the range (0,1)
    adaptive_mu_kkterror_red_fact: PositiveFloat = 0.9999
    # TODO: refine type to fix the range (0,1)
    filter_margin_fact: PositiveFloat = 1e-5
    filter_max_margin: PositiveFloat = 1
    adaptive_mu_restore_previous_iterate: YesNoBool = False
    adaptive_mu_monotone_init_factor: PositiveFloat = 0.8
    adaptive_mu_kkt_norm_type: Literal[
        "max-norm",
        "2-norm-squared",
        "1-norm",
        "2-norm",
    ] = "2-norm-squared"
    mu_strategy: Literal["monotone", "adaptive"] = "monotone"
    mu_oracle: Literal[
        "probing",
        "quality-function",
        "loqo",
    ] = "quality-function"
    fixed_mu_oracle: Literal[
        "probing",
        "loqo",
        "quality-function",
        "average_compl",
    ] = "average_compl"
    mu_init: PositiveFloat = 0.1
    barrier_tol_factor: PositiveFloat = 10
    # TODO: refine type to fix the range (0,1)
    mu_linear_decrease_factor: PositiveFloat = 0.2
    # TODO: refine type to fix the range (1,2)
    mu_superlinear_decrease_power: GtOneFloat = 1.5
    mu_allow_fast_monotone_decrease: YesNoBool = True
    # TODO: refine type to fix the range (0,1)
    tau_min: PositiveFloat = 0.99
    sigma_max: PositiveFloat = 100
    sigma_min: NonNegativeFloat = 1e-6
    quality_function_norm_type: Literal[
        "max-norm",
        "2-norm-squared",
        "1-norm",
        "2-norm",
    ] = "2-norm-squared"
    quality_function_centrality: (
        Literal[
            "none",
            "reciprocal",
            "log",
            "cubed-reciprocal",
        ]
        | None
    ) = None
    quality_function_balancing_term: Literal["none", "cubic"] | None = None
    quality_function_max_section_steps: NonNegativeInt = 8
    # TODO: refine type to fix the range [0,1)
    quality_function_section_sigma_tol: NonNegativeFloat = 0.01
    # TODO: refine type to fix the range [0,1)
    quality_function_section_qf_tol: NonNegativeFloat = 0.0
    # line search
    line_search_method: Literal[
        "filter",
        "penalty",
        "cg-penalty",
    ] = "filter"
    # TODO: refine type to fix the range (0,1)
    alpha_red_factor: PositiveFloat = 0.5
    accept_every_trial_step: YesNoBool = False
    accept_after_max_steps: Literal[-1] | NonNegativeInt = -1
    alpha_for_y: Literal[
        "primal",
        "bound-mult",
        "min",
        "max",
        "full",
        "min-dual-infeas",
        "safer-min-dual-infeas",
        "primal-and-full",
        "dual-and-full",
        "acceptor",
    ] = "primal"
    alpha_for_y_tol: NonNegativeFloat = 10
    tiny_step_tol: NonNegativeFloat = 2.22045 * 1e-15
    tiny_step_y_tol: NonNegativeFloat = 0.01
    watchdog_shortened_iter_trigger: NonNegativeInt = 10
    watchdog_trial_iter_max: PositiveInt = 3
    theta_max_fact: PositiveFloat = 10_000
    theta_min_fact: PositiveFloat = 0.0001
    # TODO: refine type to fix the range (0,0.5)
    eta_phi: PositiveFloat = 1e-8
    delta: PositiveFloat = 1
    s_phi: GtOneFloat = 2.3
    s_theta: GtOneFloat = 1.1
    # TODO: refine type to fix the range (0,1)
    gamma_phi: PositiveFloat = 1e-8
    # TODO: refine type to fix the range (0,1)
    gamma_theta: PositiveFloat = 1e-5
    # TODO: refine type to fix the range (0,1)
    alpha_min_frac: PositiveFloat = 0.05
    max_soc: NonNegativeInt = 4
    kappa_soc: PositiveFloat = 0.99
    obj_max_inc: float = 5.0
    max_filter_resets: NonNegativeInt = 5
    filter_reset_trigger: PositiveInt = 5
    corrector_type: (
        Literal[
            "none",
            "affine",
            "primal-dual",
        ]
        | None
    ) = None
    skip_corr_if_neg_curv: YesNoBool = True
    skip_corr_in_monotone_mode: YesNoBool = True
    corrector_compl_avrg_red_fact: PositiveFloat = 1
    soc_method: Literal[0, 1] = 0
    nu_init: PositiveFloat = 1e-6
    nu_inc: PositiveFloat = 0.0001
    # TODO: refine type to fix the range (0,1)
    rho: PositiveFloat = 0.1
    kappa_sigma: PositiveFloat = 1e10
    recalc_y: YesNoBool = False
    recalc_y_feas_tol: PositiveFloat = 1e-6
    slack_move: NonNegativeFloat = 1.81899 * 1e-12
    constraint_violation_norm_type: Literal[
        "1-norm",
        "2-norm",
        "max-norm",
    ] = "1-norm"
    # step calculation
    mehrotra_algorithm: YesNoBool = False
    fast_step_computation: YesNoBool = False
    min_refinement_steps: NonNegativeInt = 1
    max_refinement_steps: NonNegativeInt = 10
    residual_ratio_max: PositiveFloat = 1e-10
    residual_ratio_singular: PositiveFloat = 1e-5
    residual_improvement_factor: PositiveFloat = 1
    neg_curv_test_tol: NonNegativeFloat = 0
    neg_curv_test_reg: YesNoBool = True
    max_hessian_perturbation: PositiveFloat = 1e20
    min_hessian_perturbation: NonNegativeFloat = 1e-20
    perturb_inc_fact_first: GtOneFloat = 100
    perturb_inc_fact: GtOneFloat = 8
    # TODO: refine type to fix the range (0,1)
    perturb_dec_fact: PositiveFloat = 0.333333
    first_hessian_perturbation: PositiveFloat = 0.0001
    jacobian_regularization_value: NonNegativeFloat = 1e-8
    jacobian_regularization_exponent: NonNegativeFloat = 0.25
    perturb_always_cd: YesNoBool = False
    # restoration phase
    expect_infeasible_problem: YesNoBool = False
    expect_infeasible_problem_ctol: NonNegativeFloat = 0.001
    expect_infeasible_problem_ytol: PositiveFloat = 1e8
    start_with_resto: YesNoBool = False
    soft_resto_pderror_reduction_factor: NonNegativeFloat = 0.9999
    max_soft_resto_iters: NonNegativeInt = 10
    # TODO: refine type to fix the range [0,1)
    required_infeasibility_reduction: NonNegativeFloat = 0.9
    max_resto_iter: NonNegativeInt = 3_000_000
    evaluate_orig_obj_at_resto_trial: YesNoBool = True
    resto_penalty_parameter: PositiveFloat = 1000
    resto_proximity_weight: NonNegativeFloat = 1
    bound_mult_reset_threshold: NonNegativeFloat = 1000
    constr_mult_reset_threshold: NonNegativeFloat = 0
    resto_failure_feasibility_threshold: NonNegativeFloat | None = None
    # hessian approximation
    limited_memory_aug_solver: Literal[
        "sherman-morrison",
        "extended",
    ] = "sherman-morrison"
    limited_memory_max_history: NonNegativeInt = 6
    limited_memory_update_type: Literal[
        "bfgs",
        "sr1",
    ] = "bfgs"
    limited_memory_initialization: Literal[
        "scalar1",
        "scalar2",
        "scalar3",
        "scalar4",
        "constant",
    ] = "scalar1"
    limited_memory_init_val: PositiveFloat = 1
    limited_memory_init_val_max: PositiveFloat = 1e8
    limited_memory_init_val_min: PositiveFloat = 1e-8
    limited_memory_max_skipping: PositiveInt = 2
    limited_memory_special_for_resto: YesNoBool = False
    hessian_approximation: Literal[
        "limited-memory",
        "exact",
    ] = "limited-memory"
    hessian_approximation_space: Literal[
        "nonlinear-variables",
        "all-variables",
    ] = "nonlinear-variables"
    # linear solver
    linear_solver: Literal[
        "mumps", "ma27", "ma57", "ma77", "ma86", "ma97", "pardiso", "custom"
    ] = "mumps"
    linear_solver_options: dict[str, Any] | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if not self.algo_info.is_available:
            raise NotInstalledError(
                "The 'ipopt' algorithm requires the cyipopt package to be installed. "
                "You can it with: `conda install -c conda-forge cyipopt`."
            )
        if self.acceptable_tol <= self.convergence_ftol_rel:
            raise ValueError(
                "The acceptable tolerance must be larger than the desired tolerance."
            )
        if self.mu_strategy not in ["monotone", "adaptive"]:
            raise ValueError(
                f"Unknown barrier strategy: {self.mu_strategy}."
                " It must be 'monotone' or 'adaptive'."
            )
        if self.nlp_upper_bound_inf < 0:
            raise ValueError("nlp_upper_bound_inf should be > 0.")
        if self.nlp_lower_bound_inf > 0:
            raise ValueError("nlp_lower_bound_inf should be < 0.")
        linear_solver_options = (
            {} if self.linear_solver_options is None else self.linear_solver_options
        )
        if self.resto_failure_feasibility_threshold is None:
            resto_failure_feasibility_threshold = 1e2 * self.convergence_ftol_rel
        else:
            resto_failure_feasibility_threshold = (
                self.resto_failure_feasibility_threshold
            )

        # convert None to str none section
        linear_solver_options_with_none = [
            "ma86_scaling",
            "ma97_scaling",
            "ma97_scaling1",
            "ma97_scaling2",
            "ma97_scaling3",
            "spral_scaling",
            "spral_scaling_1",
            "spral_scaling_2",
            "spral_scaling_3",
            "linear_system_scaling",
        ]
        for key, val in linear_solver_options.items():
            if key in linear_solver_options_with_none:
                linear_solver_options[key] = _convert_none_to_str(val)
        boolean_linear_solver_options = [
            "linear_scaling_on_demand"
            "ma27_skip_inertia_check"
            "ma27_ignore_singularity"
            "ma57_automatic_scaling"
            "ma97_solve_blas3"
            "pardiso_redo_symbolic_fact_only_if_inertia_wrong"
            "pardiso_repeated_perturbation_means_singular"
            "pardiso_skip_inertia_check"
            "pardiso_iterative"
            "pardisomkl_redo_symbolic_fact_only_if_inertia_wrong"
            "pardisomkl_repeated_perturbation_means_singular"
            "pardisomkl_skip_inertia_check"
            "spral_ignore_numa"
            "spral_use_gpu"
            "wsmp_skip_inertia_check"
            "wsmp_no_pivoting"
        ]
        for key, val in linear_solver_options.items():
            if key in boolean_linear_solver_options:
                linear_solver_options[key] = _convert_bool_to_str(val, key)

        convert_bool_to_str_options = {
            "dependency_detection_with_rhs": self.dependency_detection_with_rhs,
            "check_derivatives_for_naninf": self.check_derivatives_for_naninf,
            "jac_c_constant": self.jac_c_constant,
            "jac_d_constant": self.jac_d_constant,
            "hessian_constant": self.hessian_constant,
            "least_square_init_primal": self.least_square_init_primal,
            "least_square_init_duals": self.least_square_init_duals,
            "warm_start_init_point": self.warm_start_init_point,
            "warm_start_same_structure": self.warm_start_same_structure,
            "warm_start_entire_iterate": self.warm_start_entire_iterate,
            "replace_bounds": self.replace_bounds,
            "skip_finalize_solution_call": self.skip_finalize_solution_call,
            "timing_statistics": self.timing_statistics,
            "adaptive_mu_restore_previous_iterate": (
                self.adaptive_mu_restore_previous_iterate
            ),
            "mu_allow_fast_monotone_decrease": self.mu_allow_fast_monotone_decrease,
            "accept_every_trial_step": self.accept_every_trial_step,
            "skip_corr_if_neg_curv": self.skip_corr_if_neg_curv,
            "skip_corr_in_monotone_mode": self.skip_corr_in_monotone_mode,
            "recalc_y": self.recalc_y,
            "mehrotra_algorithm": self.mehrotra_algorithm,
            "fast_step_computation": self.fast_step_computation,
            "neg_curv_test_reg": self.neg_curv_test_reg,
            "perturb_always_cd": self.perturb_always_cd,
            "expect_infeasible_problem": self.expect_infeasible_problem,
            "start_with_resto": self.start_with_resto,
            "evaluate_orig_obj_at_resto_trial": self.evaluate_orig_obj_at_resto_trial,
            "limited_memory_special_for_resto": self.limited_memory_special_for_resto,
            "honor_original_bounds": self.honor_original_bounds,
        }
        converted_bool_to_str_options = {
            key: _convert_bool_to_str(val, key)
            for key, val in convert_bool_to_str_options.items()
        }

        options = {
            # disable verbosity
            "print_level": 0,
            "ma77_print_level": -1,
            "ma86_print_level": -1,
            "ma97_print_level": -1,
            "pardiso_msglvl": 0,
            # disable derivative checker
            "derivative_test": "none",
            "s_max": float(self.s_max),
            "max_iter": self.stopping_maxiter,
            "max_wall_time": float(self.stopping_max_wall_time_seconds),
            "max_cpu_time": self.stopping_max_cpu_time,
            "dual_inf_tol": self.dual_inf_tol,
            "constr_viol_tol": self.constr_viol_tol,
            "compl_inf_tol": self.compl_inf_tol,
            # acceptable heuristic
            "acceptable_iter": self.acceptable_iter,
            "acceptable_tol": self.acceptable_tol,
            "acceptable_dual_inf_tol": self.acceptable_dual_inf_tol,
            "acceptable_constr_viol_tol": self.acceptable_constr_viol_tol,
            "acceptable_compl_inf_tol": self.acceptable_compl_inf_tol,
            "acceptable_obj_change_tol": self.acceptable_obj_change_tol,
            # bounds and more
            "diverging_iterates_tol": self.diverging_iterates_tol,
            "nlp_lower_bound_inf": self.nlp_lower_bound_inf,
            "nlp_upper_bound_inf": self.nlp_upper_bound_inf,
            "fixed_variable_treatment": self.fixed_variable_treatment,
            "dependency_detector": _convert_none_to_str(self.dependency_detector),
            "kappa_d": self.kappa_d,
            "bound_relax_factor": self.bound_relax_factor,
            "honor_original_bounds": self.honor_original_bounds,
            # scaling
            "nlp_scaling_method": _convert_none_to_str(self.nlp_scaling_method),
            "obj_scaling_factor": float(self.obj_scaling_factor),
            "nlp_scaling_max_gradient": float(self.nlp_scaling_max_gradient),
            "nlp_scaling_obj_target_gradient": float(
                self.nlp_scaling_obj_target_gradient
            ),
            "nlp_scaling_constr_target_gradient": float(
                self.nlp_scaling_constr_target_gradient
            ),
            "nlp_scaling_min_value": float(self.nlp_scaling_min_value),
            # initialization
            "bound_push": self.bound_push,
            "bound_frac": self.bound_frac,
            "slack_bound_push": self.slack_bound_push,
            "slack_bound_frac": self.slack_bound_frac,
            "constr_mult_init_max": float(self.constr_mult_init_max),
            "bound_mult_init_val": float(self.bound_mult_init_val),
            "bound_mult_init_method": self.bound_mult_init_method,
            # warm start
            "warm_start_bound_push": self.warm_start_bound_push,
            "warm_start_bound_frac": self.warm_start_bound_frac,
            "warm_start_slack_bound_push": self.warm_start_slack_bound_push,
            "warm_start_slack_bound_frac": self.warm_start_slack_bound_frac,
            "warm_start_mult_bound_push": self.warm_start_mult_bound_push,
            "warm_start_mult_init_max": self.warm_start_mult_init_max,
            "warm_start_target_mu": self.warm_start_target_mu,
            # more miscellaneous
            "option_file_name": self.option_file_name,
            # barrier parameter update
            "mu_target": float(self.mu_target),
            "mu_max_fact": float(self.mu_max_fact),
            "mu_max": float(self.mu_max),
            "mu_min": float(self.mu_min),
            "adaptive_mu_globalization": self.adaptive_mu_globalization,
            "adaptive_mu_kkterror_red_iters": self.adaptive_mu_kkterror_red_iters,
            "adaptive_mu_kkterror_red_fact": self.adaptive_mu_kkterror_red_fact,
            "filter_margin_fact": float(self.filter_margin_fact),
            "filter_max_margin": float(self.filter_max_margin),
            "adaptive_mu_monotone_init_factor": self.adaptive_mu_monotone_init_factor,
            "adaptive_mu_kkt_norm_type": self.adaptive_mu_kkt_norm_type,
            "mu_strategy": self.mu_strategy,
            "mu_oracle": self.mu_oracle,
            "fixed_mu_oracle": self.fixed_mu_oracle,
            "mu_init": self.mu_init,
            "barrier_tol_factor": float(self.barrier_tol_factor),
            "mu_linear_decrease_factor": self.mu_linear_decrease_factor,
            "mu_superlinear_decrease_power": self.mu_superlinear_decrease_power,
            "tau_min": self.tau_min,
            "sigma_max": float(self.sigma_max),
            "sigma_min": float(self.sigma_min),
            "quality_function_norm_type": self.quality_function_norm_type,
            "quality_function_centrality": _convert_none_to_str(
                self.quality_function_centrality
            ),
            "quality_function_balancing_term": _convert_none_to_str(
                self.quality_function_balancing_term
            ),
            "quality_function_max_section_steps": (
                self.quality_function_max_section_steps
            ),
            "quality_function_section_sigma_tol": (
                self.quality_function_section_sigma_tol
            ),
            "quality_function_section_qf_tol": self.quality_function_section_qf_tol,
            # linear search
            "line_search_method": self.line_search_method,
            "alpha_red_factor": self.alpha_red_factor,
            "accept_after_max_steps": self.accept_after_max_steps,
            "alpha_for_y": self.alpha_for_y,
            "alpha_for_y_tol": float(self.alpha_for_y_tol),
            "tiny_step_tol": self.tiny_step_tol,
            "tiny_step_y_tol": self.tiny_step_y_tol,
            "watchdog_shortened_iter_trigger": self.watchdog_shortened_iter_trigger,
            "watchdog_trial_iter_max": self.watchdog_trial_iter_max,
            "theta_max_fact": float(self.theta_max_fact),
            "theta_min_fact": self.theta_min_fact,
            "eta_phi": self.eta_phi,
            "delta": float(self.delta),
            "s_phi": self.s_phi,
            "s_theta": self.s_theta,
            "gamma_phi": self.gamma_phi,
            "gamma_theta": self.gamma_theta,
            "alpha_min_frac": self.alpha_min_frac,
            "max_soc": self.max_soc,
            "kappa_soc": self.kappa_soc,
            "obj_max_inc": float(self.obj_max_inc),
            "max_filter_resets": self.max_filter_resets,
            "filter_reset_trigger": self.filter_reset_trigger,
            "corrector_type": _convert_none_to_str(self.corrector_type),
            "corrector_compl_avrg_red_fact": float(self.corrector_compl_avrg_red_fact),
            "soc_method": self.soc_method,
            "nu_init": self.nu_init,
            "nu_inc": self.nu_inc,
            "rho": self.rho,
            "kappa_sigma": self.kappa_sigma,
            "recalc_y_feas_tol": self.recalc_y_feas_tol,
            "slack_move": self.slack_move,
            "constraint_violation_norm_type": self.constraint_violation_norm_type,
            # step calculation
            "min_refinement_steps": self.min_refinement_steps,
            "max_refinement_steps": self.max_refinement_steps,
            "residual_ratio_max": self.residual_ratio_max,
            "residual_ratio_singular": self.residual_ratio_singular,
            "residual_improvement_factor": float(self.residual_improvement_factor),
            "neg_curv_test_tol": float(self.neg_curv_test_tol),
            "max_hessian_perturbation": self.max_hessian_perturbation,
            "min_hessian_perturbation": self.min_hessian_perturbation,
            "perturb_inc_fact_first": float(self.perturb_inc_fact_first),
            "perturb_inc_fact": float(self.perturb_inc_fact),
            "perturb_dec_fact": float(self.perturb_dec_fact),
            "first_hessian_perturbation": float(self.first_hessian_perturbation),
            "jacobian_regularization_value": float(self.jacobian_regularization_value),
            "jacobian_regularization_exponent": float(
                self.jacobian_regularization_exponent
            ),
            # restoration phase
            "expect_infeasible_problem_ctol": self.expect_infeasible_problem_ctol,
            "expect_infeasible_problem_ytol": self.expect_infeasible_problem_ytol,
            "soft_resto_pderror_reduction_factor": (
                self.soft_resto_pderror_reduction_factor
            ),
            "max_soft_resto_iters": self.max_soft_resto_iters,
            "required_infeasibility_reduction": float(
                self.required_infeasibility_reduction
            ),
            "max_resto_iter": self.max_resto_iter,
            "resto_penalty_parameter": float(self.resto_penalty_parameter),
            "resto_proximity_weight": float(self.resto_proximity_weight),
            "bound_mult_reset_threshold": float(self.bound_mult_reset_threshold),
            "constr_mult_reset_threshold": float(self.constr_mult_reset_threshold),
            "resto_failure_feasibility_threshold": float(
                resto_failure_feasibility_threshold
            ),
            # hessian approximation
            "limited_memory_aug_solver": self.limited_memory_aug_solver,
            "limited_memory_max_history": self.limited_memory_max_history,
            "limited_memory_update_type": self.limited_memory_update_type,
            "limited_memory_initialization": self.limited_memory_initialization,
            "limited_memory_init_val": float(self.limited_memory_init_val),
            "limited_memory_init_val_max": self.limited_memory_init_val_max,
            "limited_memory_init_val_min": self.limited_memory_init_val_min,
            "limited_memory_max_skipping": self.limited_memory_max_skipping,
            "hessian_approximation": self.hessian_approximation,
            "hessian_approximation_space": self.hessian_approximation_space,
            # linear solver
            "linear_solver": self.linear_solver,
            **linear_solver_options,
            **converted_bool_to_str_options,
        }

        raw_res = cyipopt.minimize_ipopt(
            fun=problem.fun,
            x0=x0,
            bounds=_get_scipy_bounds(problem.bounds),
            jac=problem.jac,
            constraints=problem.nonlinear_constraints,
            tol=self.convergence_ftol_rel,
            options=options,
        )

        res = process_scipy_result(raw_res)

        return res


def _get_scipy_bounds(bounds: InternalBounds) -> ScipyBounds:
    return ScipyBounds(lb=bounds.lower, ub=bounds.upper)


def _convert_bool_to_str(var, name):
    """Convert input to either 'yes' or 'no' and check the output is yes or no.

    Args:
        var (str or bool): user input
        name (str): name of the variable.

    Returns:
        out (str): "yes" or "no".

    """
    if var is True:
        out = "yes"
    elif var is False:
        out = "no"
    else:
        out = var
    if out not in {"yes", "no"}:
        raise ValueError(
            f"{name} must be 'yes', 'no', True or False. You specified {var}."
        )
    return out


def _convert_none_to_str(var):
    out = "none" if var is None else var
    return out



import io
import warnings
from abc import ABC, abstractmethod
from dataclasses import asdict, fields, is_dataclass
from typing import Any, Generic, Type, TypeVar

import cloudpickle
import pandas as pd

from optimagic.exceptions import get_traceback
from optimagic.typing import DictLikeAccess

InputType = TypeVar("InputType", bound=DictLikeAccess)
OutputType = TypeVar("OutputType", bound=DictLikeAccess)


class _KeyValueStore(Generic[InputType, OutputType], ABC):
    """Generic abstract base class for a key-value store.

    This class defines the basic interface for key-value stores that support
    insertion and selection of items based on a primary key.

    Args:
        input_type: The type of input data that can be stored.
        output_type: The type of output data that can be retrieved.
        primary_key: The primary key used to uniquely identify items in the store.

    Raises:
        ValueError: If input_type or output_type is not a dataclass, or if
                the primary key is not found in output_type fields.

    """

    def __init__(
        self,
        input_type: Type[InputType],
        output_type: Type[OutputType],
        primary_key: str,
    ):
        if not (is_dataclass(input_type) and is_dataclass(output_type)):
            raise ValueError("Arguments input_type and output_type must by dataclasses")

        output_fields = {f.name for f in fields(output_type)}
        if primary_key not in output_fields:
            raise ValueError(
                f"Primary key {primary_key} not found in output_type fields "
                f"{fields(output_type)}"
            )

        self._output_type = output_type
        self._input_type = input_type
        self._primary_key = primary_key
        self._supported_fields = {f.name for f in fields(input_type)}

    @property
    def primary_key(self) -> str:
        """Get the primary key of the store.

        Returns:
            The primary key field name.

        """
        return self._primary_key

    @abstractmethod
    def insert(self, value: InputType) -> None:
        """Implement this method to insert a new value into the key-value store.

        Make sure an auto-increment logic is implemented for the insertion.

        """

    @abstractmethod
    def _select_by_key(self, key: int) -> list[OutputType]:
        """Implement this method to select a value from the store by its primary key."""

    @abstractmethod
    def _select_all(self) -> list[OutputType]:
        """Implement this method to select all values from the store."""

    def select(self, key: int | None = None) -> list[OutputType]:
        """Select items from the store.

        Args:
            key: Optional; the primary key of the item to be selected. If not provided,
                 all items will be selected.

        Returns:
            A list of output items.

        """
        if key is None:
            return self._select_all()

        return self._select_by_key(key)

    @abstractmethod
    def select_last_rows(self, n_rows: int) -> list[OutputType]:
        """Implement this to select the last `n_rows` from the store.

        Args:
            n_rows: The number of rows to select.

        Returns:
            A list of the last `n_rows` output items.

        """

    def to_df(self) -> pd.DataFrame:
        """Convert the store's data to a Pandas DataFrame.

        Returns:
            A DataFrame containing all items in the store.

        """
        items = self._select_all()
        return pd.DataFrame([asdict(item) for item in items])


class UpdatableKeyValueStore(_KeyValueStore[InputType, OutputType], ABC):
    """Generic abstract base class for an updatable key-value store.

    This class extends `KeyValueStore` to add support for updating existing
    items in the store.

    """

    def update(self, key: int, value: InputType | dict[str, Any]) -> None:
        """Update an existing item in the store.

        Args:
            key: The primary key of the item to be updated.
            value: The updated item, or a dictionary of fields to update.

        Raises:
            ValueError: If any fields in `value` are not supported by the store.

        """
        self._check_fields(value)
        self._update(key, value)

    @abstractmethod
    def _update(self, key: int, value: InputType | dict[str, Any]) -> None:
        """Implement the internal method to update an existing item in the store."""

    def _check_fields(self, value: InputType | dict[str, Any]) -> None:
        if isinstance(value, dict):
            not_supported_fields = set(value.keys()).difference(self._supported_fields)
            if not_supported_fields:
                raise ValueError(
                    f"Not supported fields {not_supported_fields}. "
                    f"Only supports fields {self._supported_fields}"
                )


class NonUpdatableKeyValueStore(_KeyValueStore[InputType, OutputType], ABC):
    def __getattr__(self, name: str) -> Any:
        if name == "update":
            msg = (
                f"'{self.__class__.__name__}' object does not allow to update items in"
                f"the store"
            )
        else:
            msg = f"'{self.__class__.__name__}' object has no attribute '{name}'"
        raise AttributeError(msg)


class RobustPickler:
    @staticmethod
    def loads(
        data: Any,
        fix_imports: bool = True,  # noqa: ARG004
        encoding: str = "ASCII",  # noqa: ARG004
        errors: str = "strict",  # noqa: ARG004
        buffers: Any = None,  # noqa: ARG004
    ) -> Any:
        """Robust pickle loading.

        We first try to unpickle the object with pd.read_pickle. This makes no
        difference for non-pandas objects but makes the de-serialization
        of pandas objects more robust across pandas versions. If that fails, we use
        cloudpickle. If that fails, we return None but do not raise an error.

        See: https://github.com/pandas-dev/pandas/issues/16474

        """
        try:
            res = pd.read_pickle(io.BytesIO(data), compression=None)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            try:
                res = cloudpickle.loads(data)
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception:
                res = None
                tb = get_traceback()
                warnings.warn(
                    f"Unable to read PickleType column from database:\n{tb}\n "
                    "The entry was replaced by None."
                )

        return res

    @staticmethod
    def dumps(
        obj: Any,
        protocol: str | None = None,
        *,
        fix_imports: bool = True,  # noqa: ARG001
        buffer_callback: Any = None,  # noqa: ARG004
    ) -> Any:
        return cloudpickle.dumps(obj, protocol=protocol)



from dataclasses import dataclass, fields
from enum import Enum
from typing import (
    Annotated,
    Any,
    Callable,
    ItemsView,
    Iterator,
    KeysView,
    Literal,
    Protocol,
    TypeVar,
    ValuesView,
)

import numpy as np
from annotated_types import Ge, Gt, Lt
from numpy._typing import NDArray

PyTree = Any
PyTreeRegistry = dict[type | str, dict[str, Callable[[Any], Any]]]
Scalar = Any

T = TypeVar("T")


class AggregationLevel(Enum):
    """Enum to specify the aggregation level of objective functions and solvers."""

    SCALAR = "scalar"
    LEAST_SQUARES = "least_squares"
    LIKELIHOOD = "likelihood"


class Direction(str, Enum):
    """Enum to specify the direction of optimization."""

    MINIMIZE = "minimize"
    MAXIMIZE = "maximize"


@dataclass(frozen=True)
class DictLikeAccess:
    """Useful base class for replacing string-based dictionaries with dataclass
    instances and keeping backward compatability regarding read access to the data
    structure."""

    def __getitem__(self, key: str) -> Any:
        if key in self.__dict__:
            return getattr(self, key)
        else:
            raise KeyError(f"{key} not found in {self.__class__.__name__}")

    def __iter__(self) -> Iterator[str]:
        return iter(self._dict_repr())

    def _dict_repr(self) -> dict[str, Any]:
        return {field.name: getattr(self, field.name) for field in fields(self)}

    def keys(self) -> KeysView[str]:
        return self._dict_repr().keys()

    def items(self) -> ItemsView[str, Any]:
        return self._dict_repr().items()

    def values(self) -> ValuesView[str]:
        return self._dict_repr().values()


@dataclass(frozen=True)
class TupleLikeAccess:
    """Useful base class for replacing tuples with dataclass instances and keeping
    backward compatability regarding read access to the data structure."""

    def __getitem__(self, index: int | slice) -> Any:
        field_values = [getattr(self, field.name) for field in fields(self)]
        return field_values[index]

    def __len__(self) -> int:
        return len(fields(self))

    def __iter__(self) -> Iterator[str]:
        for field in fields(self):
            yield getattr(self, field.name)


class ErrorHandling(Enum):
    """Enum to specify the error handling strategy of the optimization algorithm."""

    RAISE = "raise"
    RAISE_STRICT = "raise_strict"
    CONTINUE = "continue"


class EvalTask(Enum):
    """Enum to specify the task of the evaluation function."""

    FUN = "fun"
    JAC = "jac"
    FUN_AND_JAC = "fun_and_jac"
    EXPLORATION = "exploration"


class BatchEvaluator(Protocol):
    def __call__(
        self,
        func: Callable[..., T],
        arguments: list[Any],
        n_cores: int = 1,
        error_handling: ErrorHandling
        | Literal["raise", "continue"] = ErrorHandling.CONTINUE,
        unpack_symbol: Literal["*", "**"] | None = None,
    ) -> list[T]:
        pass


PositiveInt = Annotated[int, Gt(0)]
NonNegativeInt = Annotated[int, Ge(0)]
PositiveFloat = Annotated[float, Gt(0)]
NonNegativeFloat = Annotated[float, Ge(0)]
NegativeFloat = Annotated[float, Lt(0)]
GtOneFloat = Annotated[float, Gt(1)]
YesNoBool = Literal["yes", "no"] | bool
DirectionLiteral = Literal["minimize", "maximize"]
BatchEvaluatorLiteral = Literal["joblib", "pathos"]
ErrorHandlingLiteral = Literal["raise", "continue"]


@dataclass(frozen=True)
class IterationHistory(DictLikeAccess):
    """History of iterations in a process.

    Attributes:
        params: A list of parameters used in each iteration.
        criterion: A list of criterion values obtained in each iteration.
        runtime: A list or array of runtimes associated with each iteration.

    """

    params: list[PyTree]
    fun: list[float]
    time: list[float] | NDArray[np.float64]


@dataclass(frozen=True)
class MultiStartIterationHistory(TupleLikeAccess):
    """History of multiple start iterations.

    Attributes:
        history: The main iteration history, representing the best end value.
        local_histories: Optional, a list of local iteration histories.
        exploration: Optional, iteration history for exploration steps.

    """

    history: IterationHistory
    local_histories: list[IterationHistory] | None = None
    exploration: IterationHistory | None = None



"""Implement algorithms by the (Numerical Algorithms Group)[https://www.nag.com/].

The following arguments are not supported as ``algo_options``:

- ``scaling_within_bounds``
- ``init.run_in_parallel``
- ``do_logging``, ``print_progress`` and all their advanced options.

"""

import warnings
from dataclasses import dataclass
from typing import Any, Callable, Literal, cast

import numpy as np
from numpy.typing import NDArray

from optimagic import mark
from optimagic.config import IS_DFOLS_INSTALLED, IS_PYBOBYQA_INSTALLED
from optimagic.exceptions import NotInstalledError
from optimagic.optimization.algo_options import STOPPING_MAXFUN
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalOptimizationProblem,
)
from optimagic.typing import (
    AggregationLevel,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveInt,
)
from optimagic.utilities import calculate_trustregion_initial_radius

if IS_PYBOBYQA_INSTALLED:
    import pybobyqa

if IS_DFOLS_INSTALLED:
    import dfols


CONVERGENCE_MINIMAL_TRUSTREGION_RADIUS_TOLERANCE = 1e-8
"""float: Stop when the lower trust region radius falls below this value."""

CONVERGENCE_SLOW_PROGRESS = {
    "threshold_to_characterize_as_slow": 1e-8,
    "max_insufficient_improvements": None,
    "comparison_period": 5,
}
"""dict: Specification of when to terminate or reset the optimization because of only
    slow improvements. This is similar to an absolute criterion tolerance only that
    instead of a single improvement the average over several iterations must be small.

    Possible entries are:
        threshold_to_characterize_as_slow (float): Threshold whether an improvement
            is insufficient. Note: the improvement is divided by the
            ``comparison_period``.
            So this is the required average improvement per iteration over the
            comparison period.
        max_insufficient_improvements (int): Number of consecutive
            insufficient improvements before termination (or reset). Default is
            ``20 * len(x)``.
        comparison_period (int):
            How many iterations to go back to calculate the improvement.
            For example 5 would mean that each criterion evaluation is compared to the
            criterion value from 5 iterations before.

"""

THRESHOLD_FOR_SAFETY_STEP = 0.5
r"""float: Threshold for when to call the safety step (:math:`\gamma_s`).

    :math:`\text{proposed step} \leq \text{threshold_for_safety_step} \cdot
    \text{current_lower_trustregion_radius}`.

"""

CONVERGENCE_NOISE_CORRECTED_FTOL = 1.0
"""float: Stop when the evaluations on the set of interpolation points all fall within
    this factor of the noise level. The default is 1, i.e. when all evaluations are
    within the noise level. If you want to not use this criterion but still flag your
    criterion function as noisy, set this tolerance to 0.0.

    .. warning::
        Very small values, as in most other tolerances don't make sense here.

"""


TRUSTREGION_THRESHOLD_SUCCESSFUL = 0.1
"""float: Share of the predicted improvement that has to be achieved for a trust
    region iteration to count as successful.

"""

TRUSTREGION_THRESHOLD_VERY_SUCCESSFUL = 0.7
"""float: Share of predicted improvement that has to be achieved for a trust region
    iteration to count as very successful.``criterion_noisy``

"""

TRUSTREGION_SHRINKING_FACTOR_NOT_SUCCESSFUL = None
"""float: Ratio by which to shrink the upper trust region radius when realized
    improvement does not match the ``threshold_successful``. The default is 0.98
    if the criterion is noisy and 0.5 else.

"""

TRUSTREGION_EXPANSION_FACTOR_SUCCESSFUL = 2.0
r"""float: Ratio by which to expand the upper trust region radius :math:`\Delta_k`
    in very successful iterations (:math:`\gamma_{inc}` in the notation of the paper).

"""

TRUSTREGION_EXPANSION_FACTOR_VERY_SUCCESSFUL = 4.0
r"""float: Ratio of the proposed step ($\|s_k\|$) by which to expand the upper trust
    region radius (:math:`\Delta_k`) in very successful iterations
    (:math:`\overline{\gamma}_{inc}` in the notation of the paper).

"""

TRUSTREGION_SHRINKING_FACTOR_LOWER_RADIUS = None
r"""float: Ratio by which to shrink the lower trust region radius (:math:`\rho_k`)
    (:math:`\alpha_1` in the notation of the paper). Default is 0.9 if
    the criterion is noisy and 0.1 else.

"""

TRUSTREGION_SHRINKING_FACTOR_UPPER_RADIUS = None
r"""float: Ratio of the current lower trust region (:math:`\rho_k`) by which to shrink
    the upper trust region radius (:math:`\Delta_k`) when the lower one is shrunk
    (:math:`\alpha_2` in the notation of the paper). Default is 0.95 if the
    criterion is noisy and 0.5 else."""

RANDOM_DIRECTIONS_ORTHOGONAL = True
"""bool: Whether to make randomly drawn initial directions orthogonal."""


INTERPOLATION_ROUNDING_ERROR = 0.1
r"""float: Internally, all the NAG algorithms store interpolation points with respect
    to a base point :math:`x_b`; that is, we store :math:`\{y_t-x_b\}`,
    which reduces the risk of roundoff errors. We shift :math:`x_b` to :math:`x_k` when
    :math:`\text{proposed step} \leq \text{interpolation_rounding_error} \cdot
    \|x_k-x_b\|`.

"""

CLIP_CRITERION_IF_OVERFLOWING = True
"""bool: Whether to clip the criterion to avoid ``OverflowError``."""


TRUSTREGION_PRECONDITION_INTERPOLATION = True
"""bool: whether to scale the interpolation linear system to improve conditioning."""


RESET_OPTIONS = {
    "use_resets": None,
    "minimal_trustregion_radius_tolerance_scaling_at_reset": 1.0,
    "reset_type": "soft",
    "move_center_at_soft_reset": True,
    "reuse_criterion_value_at_hard_reset": True,
    "max_iterations_without_new_best_after_soft_reset": None,
    "auto_detect": True,
    "auto_detect_history": 30,
    "auto_detect_min_jacobian_increase": 0.015,
    "auto_detect_min_correlations": 0.1,
    "points_to_replace_at_soft_reset": 3,
    "max_consecutive_unsuccessful_resets": 10,
    # just bobyqa
    "max_unsuccessful_resets": None,
    "trust_region_scaling_at_unsuccessful_reset": None,
    # just dfols
    "max_interpolation_points": None,
    "n_extra_interpolation_points_per_soft_reset": 0,
    "n_extra_interpolation_points_per_hard_reset": 0,
    "n_additional_extra_points_to_replace_per_reset": 0,
}
r"""dict: Options for reseting the optimization.

    Possible entries are:

        use_resets (bool): Whether to do resets when the lower trust
            region radius (:math:`\rho_k`) reaches the stopping criterion
            (:math:`\rho_{end}`), or (optionally) when all interpolation points are
            within noise level. Default is ``True`` if the criterion is noisy.
        minimal_trustregion_radius_tolerance_scaling_at_reset (float): Factor with
            which the trust region stopping criterion is multiplied at each reset.

        reset_type (str): Whether to use "soft" or "hard" resets. Default is "soft".

        move_center_at_soft_reset (bool): Whether to move the trust region center
            ($x_k$) to the best new point evaluated in stead of keeping it constant.
        points_to_replace_at_soft_reset (int): Number of interpolation points to move
            at each soft reset.
        reuse_criterion_value_at_hard_reset (bool): Whether or not to recycle the
            criterion value at the best iterate found when performing a hard reset.
            This saves one criterion evaluation.
        max_iterations_without_new_best_after_soft_reset (int):
            The maximum number of successful steps in a given run where the new
            criterion value is worse than the best value found in previous runs before
            terminating. Default is ``max_criterion_evaluations``.
        auto_detect (bool): Whether or not to
            automatically determine when to reset. This is an additional condition
            and resets can still be triggered by small upper trust region radius, etc.
            There are two criteria used: upper trust region radius shrinkage
            (no increases over the history, more decreases than no changes) and
            changes in the model Jacobian (consistently increasing trend as measured
            by slope and correlation coefficient of the line of best fit).
        auto_detect_history (int):
            How many iterations of model changes and trust region radii to store.
        auto_detect_min_jacobian_increase (float):
            Minimum rate of increase of the Jacobian over past iterations to cause a
            reset.
        auto_detect_min_correlations (float):
            Minimum correlation of the Jacobian data set required to cause a reset.
        max_consecutive_unsuccessful_resets (int): maximum number of consecutive
            unsuccessful resets allowed (i.e. resets which did not outperform the
            best known value from earlier runs).

    Only used when using nag_bobyqa:

        max_unsuccessful_resets (int):
            number of total unsuccessful resets allowed.
            Default is 20 if ``seek_global_optimum`` and else unrestricted.
        trust_region_scaling_at_unsuccessful_reset (float): Factor by which to
            expand the initial lower trust region radius (:math:`\rho_{beg}`) after
            unsuccessful resets. Default is 1.1 if ``seek_global_optimum`` else 1.

    Only used when using nag_dfols:

        max_interpolation_points (int): Maximum allowed value of the number of
            interpolation points. This is useful if the number of interpolation points
            increases with each reset, e.g. when
            ``n_extra_interpolation_points_per_soft_reset > 0``. The default is
            ``n_interpolation_points``.
        n_extra_interpolation_points_per_soft_reset (int): Number of points to add to
            the interpolation set with each soft reset.
        n_extra_interpolation_points_per_hard_reset (int): Number of points to add to
            the interpolation set with each hard reset.
        n_additional_extra_points_to_replace_per_reset (int): This parameter modifies
            ``n_extra_points_to_replace_successful``. With each reset
            ``n_extra_points_to_replace_successful`` is increased by this number.

"""


TRUSTREGION_FAST_START_OPTIONS = {
    "min_inital_points": None,
    "method": "auto",
    "scale_of_trustregion_step_perturbation": None,
    "scale_of_jacobian_components_perturbation": 1e-2,
    # the following will be growing.full_rank.min_sing_val
    # but it not supported yet by DF-OLS.
    "floor_of_jacobian_singular_values": 1,
    "jacobian_max_condition_number": 1e8,
    "geometry_improving_steps": False,
    "safety_steps": True,
    "shrink_upper_radius_in_safety_steps": False,
    "full_geometry_improving_step": False,
    "reset_trustregion_radius_after_fast_start": False,
    "reset_min_trustregion_radius_after_fast_start": False,
    "shrinking_factor_not_successful": None,
    "n_extra_search_directions_per_iteration": 0,
}
r"""dict: Options to start the optimization while building the full trust region model.

    To activate this, set the number of interpolation points at which to evaluate the
    criterion before doing the first step, `min_initial_points`, to something smaller
    than the number of parameters.

    The following options can be specified:

        min_initial_points (int): Number of initial interpolation
            points in addition to the start point. This should only be changed to
            a value less than ``len(x)``, and only if the default setup cost
            of ``len(x) + 1`` evaluations of the criterion is impractical.
            If this is set to be less than the default, the input value of
            ``n_interpolation_points`` should be set to ``len(x)``.
            If the default is used, all the other parameters have no effect.
            Default is ``n_interpolation_points - 1``.
            If the default setup costs of the evaluations are very large, DF-OLS
            can start with less than ``len(x)`` interpolation points and add points
            to the trust region model with every iteration.
        method ("jacobian", "trustregion" or "auto"):
            When there are less interpolation points than ``len(x)`` the model is
            underdetermined. This can be fixed in two ways:
            If "jacobian", the interpolated Jacobian is perturbed to have full
            rank, allowing the trust region step to include components in the full
            search space. This is the default if
            ``len(x) \geq number of root contributions``.
            If "trustregion_step", the trust region step is perturbed by an
            orthogonal direction not yet searched. It is the default if
            ``len(x) < number of root contributions``.
        scale_of_trustregion_step_perturbation (float):
            When adding new search directions, the length of the step is the trust
            region radius multiplied by this value. The default is 0.1 if
            ``method == "trustregion"`` else 1.
        scale_of_jacobian_components_perturbation (float): Magnitude of extra
            components added to the Jacobian. Default is 1e-2.
        floor_of_jacobian_singular_values (float): Floor singular
            values of the Jacobian at this factor of the last non zero value.
            As of version 1.2.1 this option is not yet supported by DF-OLS!
        scale_of_jacobian_singular_value_floor (float):
            Floor singular values of the Jacobian at this factor of the last nonzero
            value.
        jacobian_max_condition_number (float): Cap on the condition number
            of Jacobian after applying floors to singular values
            (effectively another floor on the smallest singular value, since the
            largest singular value is fixed).
        geometry_improving_steps (bool): Whether to do geometry-improving steps in the
            trust region algorithm, as per the usual algorithm during the fast start.
        safety_steps (bool):
            Whether to perform safety steps.
        shrink_upper_radius_in_safety_steps (bool): During the fast start whether to
            shrink the upper trust region radius in safety steps.
        full_geometry_improving_step (bool): During the fast start whether to do a
            full geometry-improving step within safety steps (the same as the post fast
            start phase of the algorithm). Since this involves reducing the upper trust
            region radius, this can only be `True` if
            `shrink_upper_radius_in_safety_steps == False`.
        reset_trustregion_radius_after_fast_start (bool):
            Whether or not to reset the upper trust region radius to its initial value
            at the end of the fast start phase.
        reset_min_trustregion_radius_after_fast_start (bool):
            Whether or not to reset the minimum trust region radius
            (:math:`\rho_k`) to its initial value at the end of the fast start phase.
        shrinking_factor_not_successful (float):
            Ratio by which to shrink the trust region radius when realized
            improvement does not match the ``threshold_for_successful_iteration``
            during the fast start phase.  By default it is the same as
            ``reduction_when_not_successful``.
        n_extra_search_directions_per_iteration (int): Number of new search
            directions to add with each iteration where we do not have a full set
            of search directions. This approach is not recommended! Default is 0.

"""


@mark.minimizer(
    name="nag_dfols",
    solver_type=AggregationLevel.LEAST_SQUARES,
    is_available=IS_DFOLS_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NagDFOLS(Algorithm):
    clip_criterion_if_overflowing: bool = CLIP_CRITERION_IF_OVERFLOWING
    convergence_minimal_trustregion_radius_tolerance: NonNegativeFloat = (
        CONVERGENCE_MINIMAL_TRUSTREGION_RADIUS_TOLERANCE  # noqa: E501
    )
    convergence_noise_corrected_criterion_tolerance: NonNegativeFloat = (
        CONVERGENCE_NOISE_CORRECTED_FTOL  # noqa: E501
    )
    convergence_ftol_scaled: NonNegativeFloat = 0.0
    convergence_slow_progress: dict[str, Any] | None = None
    initial_directions: Literal[
        "coordinate",
        "random",
    ] = "coordinate"
    interpolation_rounding_error: float = INTERPOLATION_ROUNDING_ERROR
    noise_additive_level: float | None = None
    noise_multiplicative_level: float | None = None
    noise_n_evals_per_point: NonNegativeInt | None = None
    random_directions_orthogonal: bool = RANDOM_DIRECTIONS_ORTHOGONAL
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN
    threshold_for_safety_step: NonNegativeFloat = THRESHOLD_FOR_SAFETY_STEP
    trustregion_expansion_factor_successful: NonNegativeFloat = (
        TRUSTREGION_EXPANSION_FACTOR_SUCCESSFUL
    )
    trustregion_expansion_factor_very_successful: NonNegativeFloat = (
        TRUSTREGION_EXPANSION_FACTOR_VERY_SUCCESSFUL  # noqa: E501
    )
    trustregion_fast_start_options: dict[str, Any] | None = None
    trustregion_initial_radius: NonNegativeFloat | None = None
    trustregion_method_to_replace_extra_points: (
        Literal["geometry_improving", "momentum"] | None
    ) = "geometry_improving"
    trustregion_n_extra_points_to_replace_successful: NonNegativeInt = 0
    trustregion_n_interpolation_points: NonNegativeInt | None = None
    trustregion_precondition_interpolation: bool = (
        TRUSTREGION_PRECONDITION_INTERPOLATION
    )
    trustregion_reset_options: dict[str, Any] | None = None
    trustregion_shrinking_factor_not_successful: NonNegativeFloat | None = (
        TRUSTREGION_SHRINKING_FACTOR_NOT_SUCCESSFUL  # noqa: E501
    )
    trustregion_shrinking_factor_lower_radius: NonNegativeFloat | None = (
        TRUSTREGION_SHRINKING_FACTOR_LOWER_RADIUS
    )
    trustregion_shrinking_factor_upper_radius: NonNegativeFloat | None = (
        TRUSTREGION_SHRINKING_FACTOR_UPPER_RADIUS
    )
    trustregion_threshold_successful: float = TRUSTREGION_THRESHOLD_SUCCESSFUL
    trustregion_threshold_very_successful: float = TRUSTREGION_THRESHOLD_VERY_SUCCESSFUL

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = nag_dfols_internal(
            criterion=problem.fun,
            x=x0,
            lower_bounds=problem.bounds.lower,
            upper_bounds=problem.bounds.upper,
            clip_criterion_if_overflowing=self.clip_criterion_if_overflowing,
            convergence_minimal_trustregion_radius_tolerance=self.convergence_minimal_trustregion_radius_tolerance,  # noqa: E501
            convergence_noise_corrected_criterion_tolerance=self.convergence_noise_corrected_criterion_tolerance,  # noqa: E501
            convergence_ftol_scaled=self.convergence_ftol_scaled,
            convergence_slow_progress=self.convergence_slow_progress,
            initial_directions=self.initial_directions,
            interpolation_rounding_error=self.interpolation_rounding_error,
            noise_additive_level=self.noise_additive_level,
            noise_multiplicative_level=self.noise_multiplicative_level,
            noise_n_evals_per_point=self.noise_n_evals_per_point,
            random_directions_orthogonal=self.random_directions_orthogonal,
            stopping_maxfun=self.stopping_maxfun,
            threshold_for_safety_step=self.threshold_for_safety_step,
            trustregion_expansion_factor_successful=self.trustregion_expansion_factor_successful,
            trustregion_expansion_factor_very_successful=self.trustregion_expansion_factor_very_successful,  # noqa: E501
            trustregion_fast_start_options=self.trustregion_fast_start_options,
            trustregion_initial_radius=self.trustregion_initial_radius,
            trustregion_method_to_replace_extra_points=self.trustregion_method_to_replace_extra_points,
            trustregion_n_extra_points_to_replace_successful=self.trustregion_n_extra_points_to_replace_successful,
            trustregion_n_interpolation_points=self.trustregion_n_interpolation_points,
            trustregion_precondition_interpolation=self.trustregion_precondition_interpolation,
            trustregion_reset_options=self.trustregion_reset_options,
            trustregion_shrinking_factor_not_successful=self.trustregion_shrinking_factor_not_successful,
            trustregion_shrinking_factor_lower_radius=self.trustregion_shrinking_factor_lower_radius,
            trustregion_shrinking_factor_upper_radius=self.trustregion_shrinking_factor_upper_radius,
            trustregion_threshold_successful=self.trustregion_threshold_successful,
            trustregion_threshold_very_successful=self.trustregion_threshold_very_successful,
        )
        return res


def nag_dfols_internal(
    criterion,
    x,
    lower_bounds,
    upper_bounds,
    clip_criterion_if_overflowing,
    convergence_minimal_trustregion_radius_tolerance,  # noqa: E501
    convergence_noise_corrected_criterion_tolerance,  # noqa: E501
    convergence_ftol_scaled,
    convergence_slow_progress,
    initial_directions,
    interpolation_rounding_error,
    noise_additive_level,
    noise_multiplicative_level,
    noise_n_evals_per_point,
    random_directions_orthogonal,
    stopping_maxfun,
    threshold_for_safety_step,
    trustregion_expansion_factor_successful,
    trustregion_expansion_factor_very_successful,  # noqa: E501
    trustregion_fast_start_options,
    trustregion_initial_radius,
    trustregion_method_to_replace_extra_points,
    trustregion_n_extra_points_to_replace_successful,
    trustregion_n_interpolation_points,
    trustregion_precondition_interpolation,
    trustregion_reset_options,
    trustregion_shrinking_factor_not_successful,  # noqa: E501
    trustregion_shrinking_factor_lower_radius,
    trustregion_shrinking_factor_upper_radius,
    trustregion_threshold_successful,
    trustregion_threshold_very_successful,
):
    r"""Minimize a function with least squares structure using DFO-LS.

    For details see
    :ref: `list_of_nag_algorithms`.

    """
    if not IS_DFOLS_INSTALLED:
        raise NotInstalledError(
            "The 'nag_dfols' algorithm requires the DFO-LS package to be installed."
            "You can install it with 'pip install DFO-LS'. "
            "For additional installation instructions visit: ",
            r"https://numericalalgorithmsgroup.github.io/dfols/build/html/install.html",
        )
    if trustregion_method_to_replace_extra_points == "momentum":
        trustregion_use_momentum = True
    elif trustregion_method_to_replace_extra_points in ["geometry_improving", None]:
        trustregion_use_momentum = False
    else:
        raise ValueError(
            "trustregion_method_to_replace_extra_points must be "
            "'geometry_improving', 'momentum' or None."
        )

    advanced_options, trustregion_reset_options = _create_nag_advanced_options(
        x=x,
        noise_multiplicative_level=noise_multiplicative_level,
        noise_additive_level=noise_additive_level,
        noise_n_evals_per_point=noise_n_evals_per_point,
        convergence_noise_corrected_criterion_tolerance=convergence_noise_corrected_criterion_tolerance,  # noqa: E501
        trustregion_initial_radius=trustregion_initial_radius,
        trustregion_reset_options=trustregion_reset_options,
        convergence_slow_progress=convergence_slow_progress,
        interpolation_rounding_error=interpolation_rounding_error,
        threshold_for_safety_step=threshold_for_safety_step,
        clip_criterion_if_overflowing=clip_criterion_if_overflowing,
        initial_directions=initial_directions,
        random_directions_orthogonal=random_directions_orthogonal,
        trustregion_precondition_interpolation=trustregion_precondition_interpolation,
        trustregion_threshold_successful=trustregion_threshold_successful,
        trustregion_threshold_very_successful=trustregion_threshold_very_successful,
        trustregion_shrinking_factor_not_successful=trustregion_shrinking_factor_not_successful,  # noqa: E501
        trustregion_expansion_factor_successful=trustregion_expansion_factor_successful,
        trustregion_expansion_factor_very_successful=trustregion_expansion_factor_very_successful,  # noqa: E501
        trustregion_shrinking_factor_lower_radius=trustregion_shrinking_factor_lower_radius,  # noqa: E501
        trustregion_shrinking_factor_upper_radius=trustregion_shrinking_factor_upper_radius,  # noqa: E501
    )

    fast_start = _build_options_dict(
        user_input=trustregion_fast_start_options,
        default_options=TRUSTREGION_FAST_START_OPTIONS,
    )
    if fast_start["floor_of_jacobian_singular_values"] != 1:
        warnings.warn(
            "Setting the `floor_of_jacobian_singular_values` is not supported by "
            "DF-OLS as of version 1.2.1."
        )
    if (
        fast_start["shrink_upper_radius_in_safety_steps"]
        and fast_start["full_geometry_improving_step"]
    ):
        raise ValueError(
            "full_geometry_improving_step of the trustregion_fast_start_options can "
            "only be True if shrink_upper_radius_in_safety_steps is False."
        )

    (
        faststart_jac,
        faststart_step,
    ) = _get_fast_start_method(fast_start["method"])

    if (
        trustregion_reset_options["n_extra_interpolation_points_per_soft_reset"]
        < trustregion_reset_options["n_extra_interpolation_points_per_soft_reset"]
    ):
        raise ValueError(
            "In the trustregion_reset_options "
            "'n_extra_interpolation_points_per_soft_reset' must "
            "be larger or the same as 'n_extra_interpolation_points_per_hard_reset'."
        )

    dfols_options = {
        "growing.full_rank.use_full_rank_interp": faststart_jac,
        "growing.perturb_trust_region_step": faststart_step,
        "restarts.hard.use_old_rk": trustregion_reset_options[
            "reuse_criterion_value_at_hard_reset"
        ],
        "restarts.auto_detect.min_chgJ_slope": trustregion_reset_options[
            "auto_detect_min_jacobian_increase"
        ],
        "restarts.max_npt": trustregion_reset_options["max_interpolation_points"],
        "restarts.increase_npt": trustregion_reset_options[
            "n_extra_interpolation_points_per_soft_reset"
        ]
        > 0,
        "restarts.increase_npt_amt": trustregion_reset_options[
            "n_extra_interpolation_points_per_soft_reset"
        ],
        "restarts.hard.increase_ndirs_initial_amt": trustregion_reset_options[
            "n_extra_interpolation_points_per_hard_reset"
        ]
        - trustregion_reset_options["n_extra_interpolation_points_per_soft_reset"],
        "model.rel_tol": convergence_ftol_scaled,
        "regression.num_extra_steps": trustregion_n_extra_points_to_replace_successful,
        "regression.momentum_extra_steps": trustregion_use_momentum,
        "regression.increase_num_extra_steps_with_restart": trustregion_reset_options[
            "n_additional_extra_points_to_replace_per_reset"
        ],
        "growing.ndirs_initial": fast_start["min_inital_points"],
        "growing.delta_scale_new_dirns": fast_start[
            "scale_of_trustregion_step_perturbation"
        ],
        "growing.full_rank.scale_factor": fast_start[
            "scale_of_jacobian_components_perturbation"
        ],
        "growing.full_rank.svd_max_jac_cond": fast_start[
            "jacobian_max_condition_number"
        ],
        "growing.do_geom_steps": fast_start["geometry_improving_steps"],
        "growing.safety.do_safety_step": fast_start["safety_steps"],
        "growing.safety.reduce_delta": fast_start[
            "shrink_upper_radius_in_safety_steps"
        ],
        "growing.safety.full_geom_step": fast_start["full_geometry_improving_step"],
        "growing.reset_delta": fast_start["reset_trustregion_radius_after_fast_start"],
        "growing.reset_rho": fast_start[
            "reset_min_trustregion_radius_after_fast_start"
        ],
        "growing.gamma_dec": fast_start["shrinking_factor_not_successful"],
        "growing.num_new_dirns_each_iter": fast_start[
            "n_extra_search_directions_per_iteration"
        ],
        "logging.save_diagnostic_info": True,
        "logging.save_xk": True,
    }

    advanced_options.update(dfols_options)

    raw_res = dfols.solve(
        criterion,
        x0=x,
        bounds=(lower_bounds, upper_bounds),
        maxfun=stopping_maxfun,
        rhobeg=trustregion_initial_radius,
        npt=trustregion_n_interpolation_points,
        rhoend=convergence_minimal_trustregion_radius_tolerance,
        nsamples=noise_n_evals_per_point,
        objfun_has_noise=noise_additive_level or noise_multiplicative_level,
        scaling_within_bounds=False,
        do_logging=False,
        print_progress=False,
        user_params=advanced_options,
    )

    res = _process_nag_result(raw_res, len(x))
    out = InternalOptimizeResult(
        x=res["solution_x"],
        fun=res["solution_criterion"],
        success=res["success"],
        message=res["message"],
        n_iterations=res["n_iterations"],
        n_fun_evals=res["n_fun_evals"],
    )
    return out


@mark.minimizer(
    name="nag_pybobyqa",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYBOBYQA_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class NagPyBOBYQA(Algorithm):
    clip_criterion_if_overflowing: bool = CLIP_CRITERION_IF_OVERFLOWING
    convergence_minimal_trustregion_radius_tolerance: NonNegativeFloat = (
        CONVERGENCE_MINIMAL_TRUSTREGION_RADIUS_TOLERANCE  # noqa: E501
    )
    convergence_noise_corrected_criterion_tolerance: NonNegativeFloat = (
        CONVERGENCE_NOISE_CORRECTED_FTOL  # noqa: E501
    )
    convergence_criterion_value: float | None = None
    convergence_slow_progress: dict[str, Any] | None = None
    initial_directions: Literal[
        "coordinate",
        "random",
    ] = "coordinate"
    interpolation_rounding_error: float = INTERPOLATION_ROUNDING_ERROR
    noise_additive_level: float | None = None
    noise_multiplicative_level: float | None = None
    noise_n_evals_per_point: NonNegativeInt | None = None
    random_directions_orthogonal: bool = RANDOM_DIRECTIONS_ORTHOGONAL
    seek_global_optimum: bool = False
    stopping_max_criterion_evaluations: PositiveInt = STOPPING_MAXFUN
    threshold_for_safety_step: NonNegativeFloat = THRESHOLD_FOR_SAFETY_STEP
    trustregion_expansion_factor_successful: NonNegativeFloat = (
        TRUSTREGION_EXPANSION_FACTOR_SUCCESSFUL
    )
    trustregion_expansion_factor_very_successful: NonNegativeFloat = (
        TRUSTREGION_EXPANSION_FACTOR_VERY_SUCCESSFUL  # noqa: E501
    )
    trustregion_initial_radius: NonNegativeFloat | None = None
    trustregion_minimum_change_hession_for_underdetermined_interpolation: bool = True
    trustregion_n_interpolation_points: NonNegativeInt | None = None
    trustregion_precondition_interpolation: bool = (
        TRUSTREGION_PRECONDITION_INTERPOLATION
    )
    trustregion_reset_options: dict[str, Any] | None = None
    trustregion_shrinking_factor_not_successful: NonNegativeFloat | None = (
        TRUSTREGION_SHRINKING_FACTOR_NOT_SUCCESSFUL  # noqa: E501
    )
    trustregion_shrinking_factor_lower_radius: NonNegativeFloat | None = (
        TRUSTREGION_SHRINKING_FACTOR_LOWER_RADIUS
    )
    trustregion_shrinking_factor_upper_radius: NonNegativeFloat | None = (
        TRUSTREGION_SHRINKING_FACTOR_UPPER_RADIUS
    )
    trustregion_threshold_successful: float = TRUSTREGION_THRESHOLD_SUCCESSFUL
    trustregion_threshold_very_successful: float = TRUSTREGION_THRESHOLD_VERY_SUCCESSFUL

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = nag_pybobyqa_internal(
            criterion=cast(
                Callable[[NDArray[np.float64]], NDArray[np.float64]],
                problem.fun,
            ),
            x=x0,
            lower_bounds=problem.bounds.lower,
            upper_bounds=problem.bounds.upper,
            clip_criterion_if_overflowing=self.clip_criterion_if_overflowing,
            convergence_minimal_trustregion_radius_tolerance=self.convergence_minimal_trustregion_radius_tolerance,  # noqa: E501
            convergence_noise_corrected_criterion_tolerance=self.convergence_noise_corrected_criterion_tolerance,  # noqa: E501
            convergence_slow_progress=self.convergence_slow_progress,
            convergence_criterion_value=self.convergence_criterion_value,
            initial_directions=self.initial_directions,
            interpolation_rounding_error=self.interpolation_rounding_error,
            noise_additive_level=self.noise_additive_level,
            noise_multiplicative_level=self.noise_multiplicative_level,
            noise_n_evals_per_point=self.noise_n_evals_per_point,
            random_directions_orthogonal=self.random_directions_orthogonal,
            seek_global_optimum=self.seek_global_optimum,
            stopping_max_criterion_evaluations=self.stopping_max_criterion_evaluations,
            threshold_for_safety_step=self.threshold_for_safety_step,
            trustregion_expansion_factor_successful=self.trustregion_expansion_factor_successful,
            trustregion_expansion_factor_very_successful=self.trustregion_expansion_factor_very_successful,  # noqa: E501
            trustregion_initial_radius=self.trustregion_initial_radius,
            trustregion_minimum_change_hession_for_underdetermined_interpolation=self.trustregion_minimum_change_hession_for_underdetermined_interpolation,  # noqa: E501
            trustregion_n_interpolation_points=self.trustregion_n_interpolation_points,
            trustregion_precondition_interpolation=self.trustregion_precondition_interpolation,
            trustregion_reset_options=self.trustregion_reset_options,
            trustregion_shrinking_factor_not_successful=self.trustregion_shrinking_factor_not_successful,
            trustregion_shrinking_factor_lower_radius=self.trustregion_shrinking_factor_lower_radius,
            trustregion_shrinking_factor_upper_radius=self.trustregion_shrinking_factor_upper_radius,
            trustregion_threshold_successful=self.trustregion_threshold_successful,
            trustregion_threshold_very_successful=self.trustregion_threshold_very_successful,
        )
        return res


def nag_pybobyqa_internal(
    criterion,
    x,
    lower_bounds,
    upper_bounds,
    clip_criterion_if_overflowing,
    convergence_criterion_value,
    convergence_minimal_trustregion_radius_tolerance,  # noqa: E501
    convergence_noise_corrected_criterion_tolerance,  # noqa: E501
    convergence_slow_progress,
    initial_directions,
    interpolation_rounding_error,
    noise_additive_level,
    noise_multiplicative_level,
    noise_n_evals_per_point,
    random_directions_orthogonal,
    seek_global_optimum,
    stopping_max_criterion_evaluations,
    threshold_for_safety_step,
    trustregion_expansion_factor_successful,
    trustregion_expansion_factor_very_successful,  # noqa: E501
    trustregion_initial_radius,
    trustregion_minimum_change_hession_for_underdetermined_interpolation,
    trustregion_n_interpolation_points,
    trustregion_precondition_interpolation,
    trustregion_reset_options,
    trustregion_shrinking_factor_not_successful,  # noqa: E501
    trustregion_shrinking_factor_lower_radius,
    trustregion_shrinking_factor_upper_radius,
    trustregion_threshold_successful,
    trustregion_threshold_very_successful,
):
    r"""Minimize a function using the BOBYQA algorithm.

    For details see
    :ref: `list_of_nag_algorithms`.

    """
    if not IS_PYBOBYQA_INSTALLED:
        raise NotInstalledError(
            "The 'nag_pybobyqa' algorithm requires the Py-BOBYQA package to be "
            "installed. You can install it with 'pip install Py-BOBYQA'. "
            "For additional installation instructions visit: ",
            r"https://numericalalgorithmsgroup.github.io/pybobyqa/build/html/"
            "install.html",
        )

    if convergence_criterion_value is None:
        convergence_criterion_value = -np.inf

    advanced_options, trustregion_reset_options = _create_nag_advanced_options(
        x=x,
        noise_multiplicative_level=noise_multiplicative_level,
        noise_additive_level=noise_additive_level,
        trustregion_initial_radius=trustregion_initial_radius,
        noise_n_evals_per_point=noise_n_evals_per_point,
        convergence_noise_corrected_criterion_tolerance=convergence_noise_corrected_criterion_tolerance,  # noqa: E501
        trustregion_reset_options=trustregion_reset_options,
        convergence_slow_progress=convergence_slow_progress,
        interpolation_rounding_error=interpolation_rounding_error,
        threshold_for_safety_step=threshold_for_safety_step,
        clip_criterion_if_overflowing=clip_criterion_if_overflowing,
        initial_directions=initial_directions,
        random_directions_orthogonal=random_directions_orthogonal,
        trustregion_precondition_interpolation=trustregion_precondition_interpolation,
        trustregion_threshold_successful=trustregion_threshold_successful,
        trustregion_threshold_very_successful=trustregion_threshold_very_successful,
        trustregion_shrinking_factor_not_successful=trustregion_shrinking_factor_not_successful,  # noqa: E501
        trustregion_expansion_factor_successful=trustregion_expansion_factor_successful,
        trustregion_expansion_factor_very_successful=trustregion_expansion_factor_very_successful,  # noqa: E501
        trustregion_shrinking_factor_lower_radius=trustregion_shrinking_factor_lower_radius,  # noqa: E501
        trustregion_shrinking_factor_upper_radius=trustregion_shrinking_factor_upper_radius,  # noqa: E501
    )

    pybobyqa_options = {
        "model.abs_tol": convergence_criterion_value,
        "interpolation.minimum_change_hessian": trustregion_minimum_change_hession_for_underdetermined_interpolation,  # noqa: E501
        "restarts.max_unsuccessful_restarts_total": trustregion_reset_options[
            "max_unsuccessful_resets"
        ],
        "restarts.rhobeg_scale_after_unsuccessful_restart": trustregion_reset_options[
            "trust_region_scaling_at_unsuccessful_reset"
        ],
        "restarts.hard.use_old_fk": trustregion_reset_options[
            "reuse_criterion_value_at_hard_reset"
        ],
        "restarts.auto_detect.min_chg_model_slope": trustregion_reset_options[
            "auto_detect_min_jacobian_increase"
        ],
        "logging.save_diagnostic_info": True,
        "logging.save_xk": True,
    }

    advanced_options.update(pybobyqa_options)

    raw_res = pybobyqa.solve(
        criterion,
        x0=x,
        bounds=(lower_bounds, upper_bounds),
        maxfun=stopping_max_criterion_evaluations,
        rhobeg=trustregion_initial_radius,
        user_params=advanced_options,
        scaling_within_bounds=False,
        do_logging=False,
        print_progress=False,
        objfun_has_noise=noise_additive_level or noise_multiplicative_level,
        nsamples=noise_n_evals_per_point,
        npt=trustregion_n_interpolation_points,
        rhoend=convergence_minimal_trustregion_radius_tolerance,
        seek_global_minimum=seek_global_optimum,
    )

    res = _process_nag_result(raw_res, len(x))

    out = InternalOptimizeResult(
        x=res["solution_x"],
        fun=res["solution_criterion"],
        success=res["success"],
        message=res["message"],
        n_iterations=res["n_iterations"],
    )

    return out


def _process_nag_result(nag_result_obj, len_x):
    """Convert the NAG result object to our result dictionary.

    Args:
        nag_result_obj: NAG result object
        len_x (int): length of the supplied parameters, i.e. the dimensionality of the
            problem.


    Returns:
        results (dict): See :ref:`internal_optimizer_output` for details.

    """
    if hasattr(nag_result_obj, "f"):
        solution_fun = nag_result_obj.f
    else:
        solution_fun = nag_result_obj.obj

    processed = {
        "solution_criterion": solution_fun,
        "n_fun_evals": nag_result_obj.nx,
        "message": nag_result_obj.msg,
        "success": nag_result_obj.flag == nag_result_obj.EXIT_SUCCESS,
        "reached_convergence_criterion": None,
        "diagnostic_info": nag_result_obj.diagnostic_info,
    }
    try:
        n_iterations = int(nag_result_obj.diagnostic_info["iters_total"].iloc[-1])
        processed["n_iterations"] = n_iterations
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception:
        processed["n_iterations"] = None

    if hasattr(nag_result_obj, "states"):
        processed.update({"states": nag_result_obj.states})
    if hasattr(nag_result_obj, "history_params"):
        processed.update({"history_params": nag_result_obj.history_params})
    if nag_result_obj.x is not None:
        processed["solution_x"] = nag_result_obj.x
    else:
        processed["solution_x"] = np.array([np.nan] * len_x)
    return processed


def _create_nag_advanced_options(
    x,
    noise_multiplicative_level,
    noise_additive_level,
    trustregion_initial_radius,
    noise_n_evals_per_point,
    convergence_noise_corrected_criterion_tolerance,
    trustregion_reset_options,
    convergence_slow_progress,
    interpolation_rounding_error,
    threshold_for_safety_step,
    clip_criterion_if_overflowing,
    initial_directions,
    random_directions_orthogonal,
    trustregion_precondition_interpolation,
    trustregion_threshold_successful,
    trustregion_threshold_very_successful,
    trustregion_shrinking_factor_not_successful,
    trustregion_expansion_factor_successful,
    trustregion_expansion_factor_very_successful,
    trustregion_shrinking_factor_lower_radius,
    trustregion_shrinking_factor_upper_radius,
):
    if noise_multiplicative_level is not None and noise_additive_level is not None:
        raise ValueError("You cannot specify both multiplicative and additive noise.")
    if trustregion_initial_radius is None:
        trustregion_initial_radius = calculate_trustregion_initial_radius(x)
    # -np.inf as a default leads to errors when building the documentation with sphinx.
    noise_n_evals_per_point = _change_evals_per_point_interface(noise_n_evals_per_point)
    trustregion_reset_options = _build_options_dict(
        user_input=trustregion_reset_options,
        default_options=RESET_OPTIONS,
    )
    if trustregion_reset_options["reset_type"] not in ["soft", "hard"]:
        raise ValueError(
            "reset_type in the trustregion_reset_options must be soft or hard."
        )
    if initial_directions not in ["coordinate", "random"]:
        raise ValueError("inital_directions must be either 'coordinate' or 'random'.")
    convergence_slow_progress = _build_options_dict(
        user_input=convergence_slow_progress,
        default_options=CONVERGENCE_SLOW_PROGRESS,
    )

    is_noisy = bool(noise_additive_level or noise_multiplicative_level)

    advanced_options = {
        "general.rounding_error_constant": interpolation_rounding_error,
        "general.safety_step_thresh": threshold_for_safety_step,
        "general.check_objfun_for_overflow": clip_criterion_if_overflowing,
        "tr_radius.eta1": trustregion_threshold_successful,
        "tr_radius.eta2": trustregion_threshold_very_successful,
        "tr_radius.gamma_dec": trustregion_shrinking_factor_not_successful,
        "tr_radius.gamma_inc": trustregion_expansion_factor_successful,
        "tr_radius.gamma_inc_overline": trustregion_expansion_factor_very_successful,
        "tr_radius.alpha1": trustregion_shrinking_factor_lower_radius,
        "tr_radius.alpha2": trustregion_shrinking_factor_upper_radius,
        "init.random_initial_directions": initial_directions == "random",
        "init.random_directions_make_orthogonal": random_directions_orthogonal,
        "slow.thresh_for_slow": convergence_slow_progress[
            "threshold_to_characterize_as_slow"
        ],
        "slow.max_slow_iters": convergence_slow_progress[
            "max_insufficient_improvements"
        ],
        "slow.history_for_slow": convergence_slow_progress["comparison_period"],
        "noise.multiplicative_noise_level": noise_multiplicative_level,
        "noise.additive_noise_level": noise_additive_level,
        "noise.quit_on_noise_level": (
            convergence_noise_corrected_criterion_tolerance > 0
        )
        and is_noisy,
        "noise.scale_factor_for_quit": convergence_noise_corrected_criterion_tolerance,
        "interpolation.precondition": trustregion_precondition_interpolation,
        "restarts.use_restarts": trustregion_reset_options["use_resets"],
        "restarts.max_unsuccessful_restarts": trustregion_reset_options[
            "max_consecutive_unsuccessful_resets"
        ],
        "restarts.rhoend_scale": trustregion_reset_options[
            "minimal_trustregion_radius_tolerance_scaling_at_reset"
        ],
        "restarts.use_soft_restarts": trustregion_reset_options["reset_type"] == "soft",
        "restarts.soft.move_xk": trustregion_reset_options["move_center_at_soft_reset"],
        "restarts.soft.max_fake_successful_steps": trustregion_reset_options[
            "max_iterations_without_new_best_after_soft_reset"
        ],
        "restarts.auto_detect": trustregion_reset_options["auto_detect"],
        "restarts.auto_detect.history": trustregion_reset_options[
            "auto_detect_history"
        ],
        "restarts.auto_detect.min_correl": trustregion_reset_options[
            "auto_detect_min_correlations"
        ],
        "restarts.soft.num_geom_steps": trustregion_reset_options[
            "points_to_replace_at_soft_reset"
        ],
    }

    return advanced_options, trustregion_reset_options


def _change_evals_per_point_interface(func):
    """Change the interface of the user supplied function to the one expected by NAG.

    Args:
        func (callable or None): function mapping from our names to
            noise_n_evals_per_point.

    Returns:
        adjusted_noise_n_evals_per_point (callable): function mapping from the
            argument names expected by pybobyqa and df-ols to noise_n_evals_per_point.

    """
    if func is not None:

        def adjusted_noise_n_evals_per_point(delta, rho, iter, nrestarts):  # noqa: A002
            return func(
                upper_trustregion_radius=delta,
                lower_trustregion_radius=rho,
                n_iterations=iter,
                n_resets=nrestarts,
            )

        return adjusted_noise_n_evals_per_point


def _build_options_dict(user_input, default_options):
    """Create the full dictionary of trust region fast start options from user input.

    Args:
        user_input (dict or None): dictionary to update the default options with.
            May only contain keys present in the default options.
        default_options (dict): the default values.

    Returns:
        full_options (dict)

    """
    full_options = default_options.copy()
    user_input = {} if user_input is None else user_input
    invalid = [x for x in user_input if x not in full_options]
    if len(invalid) > 0:
        raise ValueError(
            f"You specified illegal options {', '.join(invalid)}. Allowed are: "
            ", ".join(full_options.keys())
        )
    full_options.update(user_input)
    return full_options


def _get_fast_start_method(user_value):
    """Get fast start method arguments from user value."""
    allowed_values = ["auto", "jacobian", "trustregion"]
    if user_value not in allowed_values:
        raise ValueError(
            "`perturb_jacobian_or_trustregion_step` must be one of "
            f"{allowed_values}. You provided {user_value}."
        )
    if user_value == "auto":
        faststart_jac = None
        faststart_step = None
    else:
        faststart_jac = user_value == "jacobian"
        faststart_step = not faststart_jac

    return faststart_jac, faststart_step



import warnings
from dataclasses import dataclass

from estimagic import utilities
from estimagic.bootstrap import BootstrapResult, bootstrap
from estimagic.estimate_ml import LikelihoodResult, estimate_ml
from estimagic.estimate_msm import MomentsResult, estimate_msm
from estimagic.estimation_table import (
    estimation_table,
    render_html,
    render_latex,
)
from estimagic.lollipop_plot import lollipop_plot
from estimagic.msm_weighting import get_moments_cov
from optimagic import OptimizeLogReader as _OptimizeLogReader
from optimagic import OptimizeResult as _OptimizeResult
from optimagic import __version__
from optimagic import check_constraints as _check_constraints
from optimagic import convergence_plot as _convergence_plot
from optimagic import convergence_report as _convergence_report
from optimagic import count_free_params as _count_free_params
from optimagic import criterion_plot as _criterion_plot
from optimagic import first_derivative as _first_derivative
from optimagic import get_benchmark_problems as _get_benchmark_problems
from optimagic import maximize as _maximize
from optimagic import minimize as _minimize
from optimagic import params_plot as _params_plot
from optimagic import profile_plot as _profile_plot
from optimagic import rank_report as _rank_report
from optimagic import run_benchmark as _run_benchmark
from optimagic import second_derivative as _second_derivative
from optimagic import slice_plot as _slice_plot
from optimagic import traceback_report as _traceback_report
from optimagic.decorators import deprecated

MSG = (
    "estimagic.{name} has been deprecated in version 0.5.0. Use optimagic.{name} "
    "instead. This function will be removed in version 0.6.0."
)

minimize = deprecated(_minimize, MSG.format(name="minimize"))
maximize = deprecated(_maximize, MSG.format(name="maximize"))
first_derivative = deprecated(_first_derivative, MSG.format(name="first_derivative"))
second_derivative = deprecated(_second_derivative, MSG.format(name="second_derivative"))
run_benchmark = deprecated(_run_benchmark, MSG.format(name="run_benchmark"))
get_benchmark_problems = deprecated(
    _get_benchmark_problems, MSG.format(name="get_benchmark_problems")
)
convergence_report = deprecated(
    _convergence_report, MSG.format(name="convergence_report")
)
rank_report = deprecated(_rank_report, MSG.format(name="rank_report"))
traceback_report = deprecated(_traceback_report, MSG.format(name="traceback_report"))
profile_plot = deprecated(_profile_plot, MSG.format(name="profile_plot"))
convergence_plot = deprecated(_convergence_plot, MSG.format(name="convergence_plot"))
slice_plot = deprecated(_slice_plot, MSG.format(name="slice_plot"))
check_constraints = deprecated(_check_constraints, MSG.format(name="check_constraints"))
count_free_params = deprecated(_count_free_params, MSG.format(name="count_free_params"))
criterion_plot = deprecated(_criterion_plot, MSG.format(name="criterion_plot"))
params_plot = deprecated(_params_plot, MSG.format(name="params_plot"))


class OptimizeLogReader(_OptimizeLogReader):
    def __init__(self, path):
        warnings.warn(
            "estimagic.OptimizeLogReader has been deprecated in version 0.5.0. Use "
            "optimagic.OptimizeLogReader instead. This class will be removed in version"
            " 0.6.0.",
            FutureWarning,
        )
        super().__init__(path)


@dataclass
class OptimizeResult(_OptimizeResult):
    def __post_init__(self):
        warnings.warn(
            "estimagic.OptimizeResult has been deprecated in version 0.5.0. Use "
            "optimagic.OptimizeResult instead. This class will be removed in version "
            "0.6.0.",
            FutureWarning,
        )


__all__ = [
    "LikelihoodResult",
    "estimate_ml",
    "estimate_msm",
    "MomentsResult",
    "estimate_msm",
    "BootstrapResult",
    "bootstrap",
    "get_moments_cov",
    "estimation_table",
    "render_html",
    "render_latex",
    "utilities",
    "minimize",
    "maximize",
    "first_derivative",
    "second_derivative",
    "run_benchmark",
    "get_benchmark_problems",
    "profile_plot",
    "convergence_plot",
    "convergence_report",
    "rank_report",
    "traceback_report",
    "lollipop_plot",
    "slice_plot",
    "check_constraints",
    "count_free_params",
    "OptimizeLogReader",
    "OptimizeResult",
    "criterion_plot",
    "params_plot",
    "__version__",
]



import typing
import warnings
from abc import ABC, ABCMeta, abstractmethod
from dataclasses import dataclass, replace
from typing import Any

import numpy as np
from numpy.typing import NDArray
from typing_extensions import Self

from optimagic.exceptions import InvalidAlgoInfoError, InvalidAlgoOptionError
from optimagic.logging.types import StepStatus
from optimagic.optimization.history import History
from optimagic.optimization.internal_optimization_problem import (
    InternalOptimizationProblem,
)
from optimagic.type_conversion import TYPE_CONVERTERS
from optimagic.typing import AggregationLevel


@dataclass(frozen=True)
class AlgoInfo:
    name: str
    solver_type: AggregationLevel
    is_available: bool
    is_global: bool
    needs_jac: bool
    needs_hess: bool
    supports_parallelism: bool
    supports_bounds: bool
    supports_linear_constraints: bool
    supports_nonlinear_constraints: bool
    disable_history: bool = False

    def __post_init__(self) -> None:
        report: list[str] = []
        if not isinstance(self.name, str):
            report.append("name must be a string")
        if not isinstance(self.solver_type, AggregationLevel):
            report.append("problem_type must be an AggregationLevel")
        if not isinstance(self.is_available, bool):
            report.append("is_available must be a bool")
        if not isinstance(self.is_global, bool):
            report.append("is_global must be a bool")
        if not isinstance(self.needs_jac, bool):
            report.append("needs_jac must be a bool")
        if not isinstance(self.needs_hess, bool):
            report.append("needs_hess must be a bool")
        if not isinstance(self.supports_parallelism, bool):
            report.append("supports_parallelism must be a bool")
        if not isinstance(self.supports_bounds, bool):
            report.append("supports_bounds must be a bool")
        if not isinstance(self.supports_linear_constraints, bool):
            report.append("supports_linear_constraints must be a bool")
        if not isinstance(self.supports_nonlinear_constraints, bool):
            report.append("supports_nonlinear_constraints must be a bool")
        if not isinstance(self.disable_history, bool):
            report.append("disable_history must be a bool")

        if report:
            msg = (
                "The following arguments to AlgoInfo or `mark.minimizer` are "
                "invalid:\n" + "\n".join(report)
            )
            raise InvalidAlgoInfoError(msg)


@dataclass(frozen=True)
class InternalOptimizeResult:
    x: NDArray[np.float64]
    fun: float | NDArray[np.float64]
    success: bool | None = None
    message: str | None = None
    status: int | None = None
    n_fun_evals: int | None = None
    n_jac_evals: int | None = None
    n_hess_evals: int | None = None
    n_iterations: int | None = None
    jac: NDArray[np.float64] | None = None
    hess: NDArray[np.float64] | None = None
    hess_inv: NDArray[np.float64] | None = None
    max_constraint_violation: float | None = None
    info: dict[str, typing.Any] | None = None
    history: History | None = None
    multistart_info: dict[str, typing.Any] | None = None

    def __post_init__(self) -> None:
        report: list[str] = []
        if not isinstance(self.x, np.ndarray):
            report.append("x must be a numpy array")

        if not (isinstance(self.fun, np.ndarray) or np.isscalar(self.fun)):
            report.append("fun must be a numpy array or scalar")

        if self.success is not None and not isinstance(self.success, bool):
            report.append("success must be a bool or None")

        if self.message is not None and not isinstance(self.message, str):
            report.append("message must be a string or None")

        if self.n_fun_evals is not None and not isinstance(self.n_fun_evals, int):
            report.append("n_fun_evals must be an int or None")

        if self.n_jac_evals is not None and not isinstance(self.n_jac_evals, int):
            report.append("n_jac_evals must be an int or None")

        if self.n_hess_evals is not None and not isinstance(self.n_hess_evals, int):
            report.append("n_hess_evals must be an int or None")

        if self.n_iterations is not None and not isinstance(self.n_iterations, int):
            report.append("n_iterations must be an int or None")

        if self.jac is not None and not isinstance(self.jac, np.ndarray):
            report.append("jac must be a numpy array or None")

        if self.hess is not None and not isinstance(self.hess, np.ndarray):
            report.append("hess must be a numpy array or None")

        if self.hess_inv is not None and not isinstance(self.hess_inv, np.ndarray):
            report.append("hess_inv must be a numpy array or None")

        if self.max_constraint_violation is not None and not np.isscalar(
            self.max_constraint_violation
        ):
            report.append("max_constraint_violation must be a scalar or None")

        if self.info is not None and not isinstance(self.info, dict):
            report.append("info must be a dictionary or None")

        if self.status is not None and not isinstance(self.status, int):
            report.append("status must be an int or None")

        if self.max_constraint_violation and not isinstance(
            self.max_constraint_violation, float
        ):
            report.append("max_constraint_violation must be a float or None")

        if report:
            msg = (
                "The following arguments to InternalOptimizeResult are invalid:\n"
                + "\n".join(report)
            )
            raise TypeError(msg)


class AlgorithmMeta(ABCMeta):
    """Metaclass to get repr, algo_info and name for classes, not just instances."""

    def __repr__(self) -> str:
        if hasattr(self, "__algo_info__") and self.__algo_info__ is not None:
            out = f"om.algos.{self.__algo_info__.name}"
        else:
            out = self.__class__.__name__
        return out

    @property
    def name(self) -> str:
        if hasattr(self, "__algo_info__") and self.__algo_info__ is not None:
            out = self.__algo_info__.name
        else:
            out = self.__class__.__name__
        return out

    @property
    def algo_info(self) -> AlgoInfo:
        if not hasattr(self, "__algo_info__") or self.__algo_info__ is None:
            msg = (
                f"The algorithm {self.name} does not have have the __algo_info__ "
                "attribute. Use the `mark.minimizer` decorator to add this attribute."
            )
            raise AttributeError(msg)

        return self.__algo_info__


@dataclass(frozen=True)
class Algorithm(ABC, metaclass=AlgorithmMeta):
    @abstractmethod
    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        pass

    def __post_init__(self) -> None:
        for field in self.__dataclass_fields__:
            raw_value = getattr(self, field)
            target_type = typing.cast(type, self.__dataclass_fields__[field].type)
            if target_type in TYPE_CONVERTERS:
                try:
                    value = TYPE_CONVERTERS[target_type](raw_value)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except Exception as e:
                    msg = (
                        f"Could not convert the value of the field {field} to the "
                        f"expected type {target_type}."
                    )
                    raise InvalidAlgoOptionError(msg) from e

                object.__setattr__(self, field, value)

    def with_option(self, **kwargs: Any) -> Self:
        valid_keys = set(self.__dataclass_fields__) - {"__algo_info__"}
        invalid = set(kwargs) - valid_keys
        if invalid:
            raise InvalidAlgoOptionError(
                f"The keyword arguments {invalid} are not valid options for "
                f"the algorithm {self.name}"
            )
        return replace(self, **kwargs)

    def with_stopping(self, **kwargs: Any) -> Self:
        options = {}
        for k, v in kwargs.items():
            if k.startswith("stopping_"):
                options[k] = v
            else:
                options[f"stopping_{k}"] = v

        return self.with_option(**options)

    def with_convergence(self, **kwargs: Any) -> Self:
        options = {}
        for k, v in kwargs.items():
            if k.startswith("convergence_"):
                options[k] = v
            else:
                options[f"convergence_{k}"] = v

        return self.with_option(**options)

    def solve_internal_problem(
        self,
        problem: InternalOptimizationProblem,
        x0: NDArray[np.float64],
        step_id: int,
    ) -> InternalOptimizeResult:
        problem = problem.with_new_history().with_step_id(step_id)

        if problem.logger:
            problem.logger.step_store.update(
                step_id, {"status": str(StepStatus.RUNNING.value)}
            )

        result = self._solve_internal_problem(problem, x0)

        if (not self.algo_info.disable_history) and (result.history is None):
            result = replace(result, history=problem.history)

        if problem.logger:
            problem.logger.step_store.update(
                step_id, {"status": str(StepStatus.COMPLETE.value)}
            )

        return result

    def with_option_if_applicable(self, **kwargs: Any) -> Self:
        """Call with_option only with applicable keyword arguments."""
        valid_keys = set(self.__dataclass_fields__) - {"__algo_info__"}
        invalid = set(kwargs) - valid_keys
        if invalid:
            msg = (
                "The following algo_options were ignored because they are not "
                f"compatible with {self.name}:\n\n {invalid}"
            )
            warnings.warn(msg)

        kwargs = {k: v for k, v in kwargs.items() if k in valid_keys}
        return self.with_option(**kwargs)

    @property
    def name(self) -> str:
        # cannot call algo_info here because it would be an infinite recursion
        if hasattr(self, "__algo_info__") and self.__algo_info__ is not None:
            return self.__algo_info__.name
        return self.__class__.__name__

    @property
    def algo_info(self) -> AlgoInfo:
        if not hasattr(self, "__algo_info__") or self.__algo_info__ is None:
            msg = (
                f"The algorithm {self.name} does not have have the __algo_info__ "
                "attribute. Use the `mark.minimizer` decorator to add this attribute."
            )
            raise AttributeError(msg)

        return self.__algo_info__



from dataclasses import dataclass
from functools import partial
from typing import Callable, Literal, Sequence, TypedDict, cast

import numpy as np
from numpy.typing import NDArray
from typing_extensions import NotRequired

from optimagic.batch_evaluators import process_batch_evaluator
from optimagic.deprecations import replace_and_warn_about_deprecated_multistart_options
from optimagic.exceptions import InvalidMultistartError
from optimagic.typing import BatchEvaluator, PyTree

# ======================================================================================
# Public Options
# ======================================================================================


@dataclass(frozen=True)
class MultistartOptions:
    """Multistart options in optimization problems.

    Attributes:
        n_samples: The number of points at which the objective function is evaluated
            during the exploration phase. If None, n_samples is set to 100 times the
            number of parameters.
        stopping_maxopt: The maximum number of local optimizations to run. Defaults to
            10% of n_samples. This number may not be reached if multistart converges
            earlier.
        sampling_distribution: The distribution from which the exploration sample is
            drawn. Allowed are "uniform" and "triangular". Defaults to "uniform".
        sampling_method: The method used to draw the exploration sample. Allowed are
            "sobol", "random", "halton", and "latin_hypercube". Defaults to "random".
        sample: A sequence of PyTrees or None. If None, a sample is drawn from the
            sampling distribution.
        mixing_weight_method: The method used to determine the mixing weight, i,e, how
            start parameters for local optimizations are calculated. Allowed are
            "tiktak" and "linear", or a custom callable. Defaults to "tiktak".
        mixing_weight_bounds: The lower and upper bounds for the mixing weight.
            Defaults to (0.1, 0.995).
        convergence_max_discoveries: The maximum number of discoveries for convergence.
            Determines after how many re-descoveries of the currently best local
            optima the multistart algorithm stops. Defaults to 2.
        convergence_xtol_rel: The relative tolerance in parameters
            for convergence. Determines the maximum relative distance two parameter
            vecctors can have to be considered equal. Defaults to 0.01.
        n_cores: The number of cores to use for parallelization. Defaults to 1.
        batch_evaluator: The evaluator to use for batch evaluation. Allowed are "joblib"
            and "pathos", or a custom callable.
        batch_size: The batch size for batch evaluation. Must be larger than n_cores
            or None.
        seed: The seed for the random number generator.
        error_handling: The error handling for exploration and optimization errors.
            Allowed are "raise" and "continue".

    Raises:
        InvalidMultistartError: If the multistart options cannot be processed, e.g.
            because they do not have the correct type.

    """

    n_samples: int | None = None
    stopping_maxopt: int | None = None
    sampling_distribution: Literal["uniform", "triangular"] = "uniform"
    sampling_method: Literal["sobol", "random", "halton", "latin_hypercube"] = "random"
    sample: Sequence[PyTree] | None = None
    mixing_weight_method: (
        Literal["tiktak", "linear"] | Callable[[int, int, float, float], float]
    ) = "tiktak"
    mixing_weight_bounds: tuple[float, float] = (0.1, 0.995)
    convergence_xtol_rel: float | None = None
    convergence_max_discoveries: int = 2
    n_cores: int = 1
    batch_evaluator: Literal["joblib", "pathos"] | BatchEvaluator = "joblib"
    batch_size: int | None = None
    seed: int | np.random.Generator | None = None
    error_handling: Literal["raise", "continue"] | None = None
    # Deprecated attributes
    share_optimization: float | None = None
    convergence_relative_params_tolerance: float | None = None
    optimization_error_handling: Literal["raise", "continue"] | None = None
    exploration_error_handling: Literal["raise", "continue"] | None = None

    def __post_init__(self) -> None:
        _validate_attribute_types_and_values(self)


class MultistartOptionsDict(TypedDict):
    n_samples: NotRequired[int | None]
    stopping_maxopt: NotRequired[int | None]
    sampling_distribution: NotRequired[Literal["uniform", "triangular"]]
    sampling_method: NotRequired[
        Literal["sobol", "random", "halton", "latin_hypercube"]
    ]
    sample: NotRequired[Sequence[PyTree] | None]
    mixing_weight_method: NotRequired[
        Literal["tiktak", "linear"] | Callable[[int, int, float, float], float]
    ]
    mixing_weight_bounds: NotRequired[tuple[float, float]]
    convergence_xtol_rel: NotRequired[float | None]
    convergence_max_discoveries: NotRequired[int]
    n_cores: NotRequired[int]
    batch_evaluator: NotRequired[Literal["joblib", "pathos"] | BatchEvaluator]
    batch_size: NotRequired[int | None]
    seed: NotRequired[int | np.random.Generator | None]
    error_handling: NotRequired[Literal["raise", "continue"] | None]
    # Deprecated attributes
    share_optimization: NotRequired[float | None]
    convergence_relative_params_tolerance: NotRequired[float | None]
    optimization_error_handling: NotRequired[Literal["raise", "continue"] | None]
    exploration_error_handling: NotRequired[Literal["raise", "continue"] | None]


def pre_process_multistart(
    multistart: bool | MultistartOptions | MultistartOptionsDict | None,
) -> MultistartOptions | None:
    """Convert all valid types of multistart to a optimagic.MultistartOptions.

    This just harmonizes multiple ways of specifying multistart options into a single
    format. It performs runime type checks, but it does not check whether multistart
    options are consistent with other option choices.

    Args:
        multistart: The user provided multistart options.
        n_params: The number of parameters in the optimization problem.

    Returns:
        The multistart options in the optimagic format.

    Raises:
        InvalidMultistartError: If the multistart options cannot be processed, e.g.
            because they do not have the correct type.

    """
    if isinstance(multistart, bool):
        multistart = MultistartOptions() if multistart else None
    elif isinstance(multistart, MultistartOptions) or multistart is None:
        pass
    else:
        try:
            multistart = MultistartOptions(**multistart)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            if isinstance(e, InvalidMultistartError):
                raise e
            raise InvalidMultistartError(
                f"Invalid multistart options of type: {type(multistart)}. Multistart "
                "options must be of type optimagic.MultistartOptions, a dictionary "
                "with valid keys, None, or a boolean."
            ) from e

    if multistart is not None:
        multistart = replace_and_warn_about_deprecated_multistart_options(multistart)
        # The replace and warn function cannot be typed due to circular imports, but
        # we know that the return type is MultistartOptions
        multistart = cast(MultistartOptions, multistart)

    return multistart


def _validate_attribute_types_and_values(options: MultistartOptions) -> None:
    if options.n_samples is not None and (
        not isinstance(options.n_samples, int) or options.n_samples < 1
    ):
        raise InvalidMultistartError(
            f"Invalid number of samples: {options.n_samples}. Number of samples "
            "must be a positive integer or None."
        )

    if options.stopping_maxopt is not None and (
        not isinstance(options.stopping_maxopt, int) or options.stopping_maxopt < 0
    ):
        raise InvalidMultistartError(
            f"Invalid number of optimizations: {options.stopping_maxopt}. Number of "
            "optimizations must be a positive integer or None."
        )

    if (
        options.n_samples is not None
        and options.stopping_maxopt is not None
        and options.n_samples < options.stopping_maxopt
    ):
        raise InvalidMultistartError(
            f"Invalid number of samples: {options.n_samples}. Number of samples "
            "must be at least as large as the number of optimizations."
        )

    if options.sampling_distribution not in ("uniform", "triangular"):
        raise InvalidMultistartError(
            f"Invalid sampling distribution: {options.sampling_distribution}. Sampling "
            f"distribution must be one of ('uniform', 'triangular')."
        )

    if options.sampling_method not in ("sobol", "random", "halton", "latin_hypercube"):
        raise InvalidMultistartError(
            f"Invalid sampling method: {options.sampling_method}. Sampling method "
            f"must be one of ('sobol', 'random', 'halton', 'latin_hypercube')."
        )

    if not isinstance(options.sample, Sequence | None):
        raise InvalidMultistartError(
            f"Invalid sample: {options.sample}. Sample must be a sequence of "
            "parameters."
        )

    if not callable(
        options.mixing_weight_method
    ) and options.mixing_weight_method not in ("tiktak", "linear"):
        raise InvalidMultistartError(
            f"Invalid mixing weight method: {options.mixing_weight_method}. Mixing "
            "weight method must be Callable or one of ('tiktak', 'linear')."
        )

    if (
        not isinstance(options.mixing_weight_bounds, tuple)
        or len(options.mixing_weight_bounds) != 2
        or not set(type(x) for x in options.mixing_weight_bounds) <= {int, float}
    ):
        raise InvalidMultistartError(
            f"Invalid mixing weight bounds: {options.mixing_weight_bounds}. Mixing "
            "weight bounds must be a tuple of two numbers."
        )

    if options.convergence_xtol_rel is not None and (
        not isinstance(options.convergence_xtol_rel, int | float)
        or options.convergence_xtol_rel < 0
    ):
        raise InvalidMultistartError(
            "Invalid relative params tolerance:"
            f"{options.convergence_xtol_rel}. Relative params "
            "tolerance must be a number."
        )

    if (
        not isinstance(options.convergence_max_discoveries, int | float)
        or options.convergence_max_discoveries < 1
    ):
        raise InvalidMultistartError(
            f"Invalid max discoveries: {options.convergence_max_discoveries}. Max "
            "discoveries must be a positive integer or infinity."
        )

    if not isinstance(options.n_cores, int) or options.n_cores < 1:
        raise InvalidMultistartError(
            f"Invalid number of cores: {options.n_cores}. Number of cores "
            "must be a positive integer."
        )

    if not callable(options.batch_evaluator) and options.batch_evaluator not in (
        "joblib",
        "pathos",
    ):
        raise InvalidMultistartError(
            f"Invalid batch evaluator: {options.batch_evaluator}. Batch evaluator "
            "must be a Callable or one of 'joblib' or 'pathos'."
        )

    if options.batch_size is not None and (
        not isinstance(options.batch_size, int) or options.batch_size < options.n_cores
    ):
        raise InvalidMultistartError(
            f"Invalid batch size: {options.batch_size}. Batch size "
            "must be a positive integer larger than n_cores, or None."
        )

    if not isinstance(options.seed, int | np.random.Generator | None):
        raise InvalidMultistartError(
            f"Invalid seed: {options.seed}. Seed "
            "must be an integer, a numpy random generator, or None."
        )

    if options.error_handling is not None and options.error_handling not in (
        "raise",
        "continue",
    ):
        raise InvalidMultistartError(
            f"Invalid error handling: {options.error_handling}. Error handling must be "
            "'raise' or 'continue'."
        )


# ======================================================================================
# Internal Options
# ======================================================================================


def _tiktak_weights(
    iteration: int, n_iterations: int, min_weight: float, max_weight: float
) -> float:
    return np.clip(np.sqrt(iteration / n_iterations), min_weight, max_weight)


def _linear_weights(
    iteration: int, n_iterations: int, min_weight: float, max_weight: float
) -> float:
    unscaled = iteration / n_iterations
    span = max_weight - min_weight
    return min_weight + unscaled * span


WEIGHT_FUNCTIONS = {
    "tiktak": _tiktak_weights,
    "linear": _linear_weights,
}


@dataclass(frozen=True)
class InternalMultistartOptions:
    """Multistart options used internally in optimagic.

    Compared to `MultistartOptions`, this data class has stricter types and combines
    some of the attributes. It is generated at runtime using a `MultistartOptions`
    instance and the function `get_internal_multistart_options_from_public`.

    """

    n_samples: int
    weight_func: Callable[[int, int], float]
    convergence_xtol_rel: float
    convergence_max_discoveries: int
    sampling_distribution: Literal["uniform", "triangular"]
    sampling_method: Literal["sobol", "random", "halton", "latin_hypercube"]
    sample: NDArray[np.float64] | None
    seed: int | np.random.Generator | None
    n_cores: int
    batch_evaluator: BatchEvaluator
    batch_size: int
    error_handling: Literal["raise", "continue"]
    stopping_maxopt: int

    def __post_init__(self) -> None:
        must_be_at_least_1 = [
            "n_samples",
            "stopping_maxopt",
            "n_cores",
            "batch_size",
            "convergence_max_discoveries",
        ]

        for attr in must_be_at_least_1:
            if getattr(self, attr) < 1:
                raise InvalidMultistartError(f"{attr} must be at least 1.")

        if self.batch_size < self.n_cores:
            raise InvalidMultistartError("batch_size must be at least n_cores.")

        if self.convergence_xtol_rel < 0:
            raise InvalidMultistartError("convergence_xtol_rel must be at least 0.")


def get_internal_multistart_options_from_public(
    options: MultistartOptions,
    params: PyTree,
    params_to_internal: Callable[[PyTree], NDArray[np.float64]],
) -> InternalMultistartOptions:
    """Get internal multistart options from public multistart options.

    Args:
        options: The pre-processed multistart options.
        params: The parameters of the optimization problem.
        params_to_internal: A function that converts parameters to internal parameters.

    Returns:
        InternalMultistartOptions: The updated options with runtime defaults.

    """
    x = params_to_internal(params)

    if options.sample is not None:
        sample = np.array([params_to_internal(x) for x in list(options.sample)])
        n_samples = len(options.sample)
    else:
        sample = None
        n_samples = options.n_samples  # type: ignore

    batch_size = options.n_cores if options.batch_size is None else options.batch_size
    batch_evaluator = process_batch_evaluator(options.batch_evaluator)

    if callable(options.mixing_weight_method):
        weight_func = options.mixing_weight_method
    else:
        _weight_method = WEIGHT_FUNCTIONS[options.mixing_weight_method]

    weight_func = partial(
        _weight_method,
        min_weight=options.mixing_weight_bounds[0],
        max_weight=options.mixing_weight_bounds[1],
    )

    if n_samples is None:
        if options.stopping_maxopt is None:
            n_samples = 100 * len(x)
        else:
            n_samples = 10 * options.stopping_maxopt

    if options.share_optimization is None:
        share_optimization = 0.1
    else:
        share_optimization = options.share_optimization

    if options.stopping_maxopt is None:
        stopping_maxopt = max(1, int(share_optimization * n_samples))
    else:
        stopping_maxopt = options.stopping_maxopt

    # Set defaults resulting from deprecated attributes
    if options.error_handling is not None:
        error_handling = options.error_handling
    else:
        error_handling = "continue"

    if options.convergence_xtol_rel is not None:
        convergence_xtol_rel = options.convergence_xtol_rel
    else:
        convergence_xtol_rel = 0.01

    return InternalMultistartOptions(
        # Attributes taken directly from MultistartOptions
        convergence_max_discoveries=options.convergence_max_discoveries,
        n_cores=options.n_cores,
        sampling_distribution=options.sampling_distribution,
        sampling_method=options.sampling_method,
        seed=options.seed,
        # Updated attributes
        sample=sample,
        n_samples=n_samples,
        weight_func=weight_func,
        error_handling=error_handling,
        convergence_xtol_rel=convergence_xtol_rel,
        stopping_maxopt=stopping_maxopt,
        batch_evaluator=batch_evaluator,
        batch_size=batch_size,
    )



from __future__ import annotations

import traceback
import warnings
from dataclasses import asdict, dataclass
from functools import cached_property
from typing import Any, Sequence, Type, cast

import sqlalchemy as sql
from sqlalchemy import Column, Integer, PickleType, String
from sqlalchemy.engine.base import Engine
from sqlalchemy.sql.base import Executable
from sqlalchemy.sql.schema import MetaData

from optimagic.logging.base import (
    InputType,
    NonUpdatableKeyValueStore,
    OutputType,
    RobustPickler,
    UpdatableKeyValueStore,
)
from optimagic.logging.types import (
    IterationState,
    IterationStateWithId,
    ProblemInitialization,
    ProblemInitializationWithId,
    StepResult,
    StepResultWithId,
)


class SQLAlchemyConfig:
    """Configuration class for setting up an SQLAlchemy engine and metadata.

    This class manages the connection URL, engine creation, and metadata reflection
    for an SQLAlchemy database connection.

    Args:
        url: The database URL to connect to.

    """

    def __init__(
        self,
        url: str,
    ):
        self.url = url

    @cached_property
    def metadata(self) -> MetaData:
        """Get the metadata object.

        Returns:
            The SQLAlchemy MetaData object reflecting the database schema.

        """
        engine = self.create_engine()
        metadata = MetaData()
        self._configure_reflect()
        metadata.reflect(engine)
        return metadata

    def create_engine(self) -> Engine:
        """Create and return an SQLAlchemy engine.

        Returns:
            An SQLAlchemy Engine object.

        """
        return sql.create_engine(self.url)

    @staticmethod
    def _configure_reflect() -> None:
        """Mark all BLOB dtypes as PickleType with our custom pickle reader.

        Code ist taken from the documentation: https://tinyurl.com/y7q287jr

        """

        @sql.event.listens_for(sql.Table, "column_reflect")
        def _setup_pickletype(
            inspector: Any, table: sql.Table, column_info: dict[str, Any]
        ) -> None:  # noqa: ARG001
            if isinstance(column_info["type"], sql.BLOB):
                column_info["type"] = sql.PickleType(pickler=RobustPickler)  # type:ignore


@dataclass
class TableConfig:
    """Configuration for creating and managing SQLAlchemy tables.

    This class defines the schema for an SQLAlchemy table, including its name,
    columns, primary key, and strategy for handling existing tables.

    Args:
        table_name: The name of the table.
        columns: A list of SQLAlchemy Column objects defining the table schema.
        primary_key: The name of the primary key column.

    """

    table_name: str
    columns: list[sql.Column[Any]]
    primary_key: str

    @property
    def column_names(self) -> list[str]:
        return [c.name for c in self.columns]

    def create_table(self, metadata: MetaData, engine: Engine) -> sql.Table:
        """Create or reflect the table in the database.

        Args:
            metadata: The SQLAlchemy MetaData object.
            engine: The SQLAlchemy Engine object.

        Returns:
            The SQLAlchemy Table object representing the created or reflected table.

        """
        metadata.reflect(engine)
        table = sql.Table(
            self.table_name, metadata, *self.columns, extend_existing=True
        )
        metadata.create_all(engine)
        return table


class _SQLAlchemyStoreMixin:
    """Mixin class for common SQLAlchemy store operations.

    This class provides common methods for selecting, inserting, and executing
    SQL statements in an SQLAlchemy-based key-value store.

    Args:
        db_config: The SQLAlchemyConfig object for database configuration.
        table_config: The TableConfig object for table configuration.

    """

    def __init__(self, db_config: SQLAlchemyConfig, table_config: TableConfig):
        self._db_config = db_config
        self._engine = db_config.create_engine()
        self._table_config = table_config
        self._table = table_config.create_table(db_config.metadata, self._engine)

    @property
    def column_names(self) -> list[str]:
        return self._table_config.column_names

    @property
    def table_name(self) -> str:
        return self._table_config.table_name

    @property
    def table(self) -> sql.Table:
        return self._table

    @property
    def engine(self) -> Engine:
        return self._engine

    def _select_row_by_key(self, key: int) -> list[Any]:
        stmt = self._table.select().where(
            getattr(self._table.c, self._table_config.primary_key) == key
        )
        return self._execute_read_statement(stmt)

    def _select_all_rows(self) -> list[Any]:
        stmt = self._table.select()
        return self._execute_read_statement(stmt)

    def _select_last_rows(self, n_rows: int) -> list[Any]:
        stmt = (
            self._table.select()
            .order_by(getattr(self._table.c, self._table_config.primary_key).desc())
            .limit(n_rows)
        )
        result = self._execute_read_statement(stmt)
        return result[::-1]

    def _insert(self, insert_values: dict[str, Any]) -> None:
        stmt = self._table.insert().values(**insert_values)
        self._execute_write_statement(stmt)

    def _execute_read_statement(self, statement: Executable) -> list[Any]:
        with self._engine.connect() as connection:
            return connection.execute(statement).fetchall()

    def _execute_write_statement(self, statement: Executable) -> None:
        try:
            with self._engine.begin() as connection:
                connection.execute(statement)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            exception_info = traceback.format_exc()
            warnings.warn(
                f"Unable to write to database. The traceback was:\n\n{exception_info}"
            )


class SQLAlchemySimpleStore(
    NonUpdatableKeyValueStore[InputType, OutputType],
    _SQLAlchemyStoreMixin,
):
    """A simple SQLAlchemy-based key-value store that does not support updates.

    This class provides basic key-value storage functionality using SQLAlchemy,
    where values are serialized and stored as BLOBs. The store does not support
    updating existing entries.

    Args:
            table_name: The name of the table.
            primary_key: The primary key column name.
            db_config: The SQLAlchemyConfig object for database configuration.

    """

    _value_column: str = "serialized_value"

    def __init__(
        self,
        table_name: str,
        primary_key: str,
        db_config: SQLAlchemyConfig,
        input_type: Type[InputType],
        output_type: Type[OutputType],
    ):
        super().__init__(input_type, output_type, primary_key)
        columns = [
            sql.Column(primary_key, sql.Integer, primary_key=True, autoincrement=True),
            sql.Column(self._value_column, sql.PickleType(pickler=RobustPickler)),  # type:ignore
        ]
        table_config = TableConfig(table_name, columns, self.primary_key)

        _SQLAlchemyStoreMixin.__init__(self, db_config, table_config)

    def __reduce__(
        self,
    ) -> tuple[
        Type[SQLAlchemySimpleStore[Any, Any]],
        tuple[str, str, SQLAlchemyConfig, Type[Any], Type[Any]],
    ]:
        return SQLAlchemySimpleStore, (
            self.table_name,
            self.primary_key,
            self._db_config,
            self._input_type,
            self._output_type,
        )

    def insert(self, value: InputType) -> None:
        """Insert a new value into the store.

        Args:
            value: The value to insert into the store.

        """
        self._insert({self._value_column: value})

    def _select_by_key(self, key: int) -> list[OutputType]:
        result = self._select_row_by_key(key)
        return self._post_process(result)

    def _select_all(self) -> list[OutputType]:
        result = self._select_all_rows()
        return self._post_process(result)

    def select_last_rows(self, n_rows: int) -> list[OutputType]:
        """Select the last `n_rows` values from the store.

        Args:
            n_rows: The number of rows to select.

        Returns:
            A list of the last `n_rows` output values.

        """
        result = self._select_last_rows(n_rows)
        return self._post_process(result)

    def _post_process(self, results: Sequence[sql.Row]) -> list[OutputType]:  # type:ignore
        output_list = []
        for row in results:
            row_dict = {self.primary_key: row[0]}
            row_dict.update(asdict(row[-1]))
            output_list.append(self._output_type(**row_dict))
        return output_list


class SQLAlchemyTableStore(
    UpdatableKeyValueStore[InputType, OutputType], _SQLAlchemyStoreMixin
):
    """An SQLAlchemy-based key-value store that supports updates.

    This class provides key-value storage functionality using SQLAlchemy,
    allowing for insertion, updating, and selection of data.

    Args:
        table_config: The TableConfig object defining the table schema.
        db_config: The SQLAlchemyConfig object for database configuration.
        input_type: The type of input data.
        output_type: The type of output data.

    """

    def __init__(
        self,
        table_config: TableConfig,
        db_config: SQLAlchemyConfig,
        input_type: Type[InputType],
        output_type: Type[OutputType],
    ):
        _SQLAlchemyStoreMixin.__init__(self, db_config, table_config)
        super().__init__(input_type, output_type, self._table_config.primary_key)

    def __reduce__(
        self,
    ) -> tuple[
        Type[SQLAlchemyTableStore[Any, Any]],
        tuple[TableConfig, SQLAlchemyConfig, Type[Any], Type[Any]],
    ]:
        return SQLAlchemyTableStore, (
            self._table_config,
            self._db_config,
            self._input_type,
            self._output_type,
        )

    def insert(self, value: InputType) -> None:
        """Insert a new value into the store.

        Args:
            value: The value to insert into the store.

        """
        self._insert(asdict(value))

    def _update(self, key: int, value: InputType | dict[str, Any]) -> None:
        if not isinstance(value, dict):
            update_values = asdict(value)
        else:
            update_values = value
        stmt = (
            self._table.update()
            .where(getattr(self._table.c, self.primary_key) == key)
            .values(**update_values)
        )
        self._execute_write_statement(stmt)

    def _select_by_key(self, key: int) -> list[OutputType]:
        result = self._select_row_by_key(key)
        return self._post_process(result)

    def _select_all(self) -> list[OutputType]:
        result = self._select_all_rows()
        return self._post_process(result)

    def select_last_rows(self, n_rows: int) -> list[OutputType]:
        """Select the last `n_rows` values from the store.

        Args:
            n_rows: The number of rows to select.

        Returns:
            A list of the last `n_rows` output values.

        """
        result = self._select_last_rows(n_rows)
        return self._post_process(result)

    def _post_process(self, results: Sequence[sql.Row]) -> list[OutputType]:  # type:ignore
        return [
            self._output_type(**dict(zip(self.column_names, row, strict=False)))
            for row in results
        ]


class IterationStore(SQLAlchemySimpleStore[IterationState, IterationStateWithId]):
    """Store for managing iteration data in an SQLite database.

    Args:
        db_config (SQLiteConfig): The SQLiteConfig object for database configuration.

    """

    _TABLE_NAME = "optimization_iterations"
    _PRIMARY_KEY = "rowid"

    def __init__(
        self,
        db_config: SQLAlchemyConfig,
    ):
        super().__init__(
            self._TABLE_NAME,
            self._PRIMARY_KEY,
            db_config,
            IterationState,
            IterationStateWithId,
        )


class StepStore(SQLAlchemyTableStore[StepResult, StepResultWithId]):
    """Store for managing step data in an SQLite database.

    Args:
        db_config (SQLiteConfig): The SQLiteConfig object for database configuration.

    """

    _TABLE_NAME = "steps"
    _PRIMARY_KEY = "rowid"

    def __init__(
        self,
        db_config: SQLAlchemyConfig,
    ):
        columns = [
            Column(self._PRIMARY_KEY, Integer, primary_key=True, autoincrement=True),
            Column("type", String),  # e.g. optimization
            Column("status", String),  # e.g. running
            Column("n_iterations", Integer),  # optional
            Column("name", String),  # e.g. "optimization-1", "exploration", not unique
        ]

        table_config = TableConfig(
            self._TABLE_NAME,
            cast(list[Column[Any]], columns),
            self._PRIMARY_KEY,
        )

        super().__init__(
            table_config,
            db_config,
            StepResult,
            StepResultWithId,
        )


class ProblemStore(
    SQLAlchemyTableStore[ProblemInitialization, ProblemInitializationWithId]
):
    """Store for managing optimization problem initialization data in an SQLite
    database.

    Args:
        db_config (SQLiteConfig): The SQLiteConfig object for database configuration.

    """

    _TABLE_NAME = "optimization_problem"
    _PRIMARY_KEY = "rowid"

    def __init__(
        self,
        db_config: SQLAlchemyConfig,
    ):
        columns = [
            Column(self._PRIMARY_KEY, Integer, primary_key=True, autoincrement=True),
            Column("direction", String),
            Column("params", PickleType(pickler=RobustPickler)),  # type:ignore
        ]

        table_config = TableConfig(
            self._TABLE_NAME,
            cast(list[Column[Any]], columns),
            self._PRIMARY_KEY,
        )

        super().__init__(
            table_config,
            db_config,
            ProblemInitialization,
            ProblemInitializationWithId,
        )



"""Implement Berndt-Hall-Hall-Hausman (BHHH) algorithm."""

from dataclasses import dataclass
from typing import Callable, cast

import numpy as np
from numpy.typing import NDArray

from optimagic import mark
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalOptimizationProblem,
)
from optimagic.typing import AggregationLevel, NonNegativeFloat, PositiveInt


@mark.minimizer(
    name="bhhh",
    solver_type=AggregationLevel.LIKELIHOOD,
    is_available=True,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=False,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class BHHH(Algorithm):
    converence_gtol_abs: NonNegativeFloat = 1e-8
    # TODO: Why is this 200?
    stopping_maxiter: PositiveInt = 200

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = bhhh_internal(
            fun_and_jac=cast(
                Callable[[NDArray[np.float64]], NDArray[np.float64]],
                problem.fun_and_jac,
            ),
            x=x0,
            gtol_abs=self.converence_gtol_abs,
            maxiter=self.stopping_maxiter,
        )

        return res


def bhhh_internal(
    fun_and_jac: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    x: NDArray[np.float64],
    gtol_abs: NonNegativeFloat,
    maxiter: PositiveInt,
) -> InternalOptimizeResult:
    """Minimize a likelihood function using the BHHH algorithm.

    Args:
        criterion_and_derivative: The objective function to be minimized.
        x: Initial guess of the parameter vector (starting points).
        convergence_absolute_gradient_tolerance: Stopping criterion for the
            gradient tolerance.
        stopping_max_iterations: Maximum number of iterations. If reached,
            terminate.

    Returns:
        InternalOptimizeResult: The result of the optimization.

    """
    criterion_accepted, gradient = fun_and_jac(x)
    x_accepted = x

    hessian_approx = np.dot(gradient.T, gradient)
    gradient_sum = np.sum(gradient, axis=0)
    direction = np.linalg.solve(hessian_approx, gradient_sum)
    gtol = np.dot(gradient_sum, direction)

    initial_step_size = 1.0
    step_size = initial_step_size

    niter = 1
    while niter < maxiter:
        niter += 1

        x_candidate = x_accepted + step_size * direction
        criterion_candidate, gradient = fun_and_jac(x_candidate)

        # If previous step was accepted
        if step_size == initial_step_size:
            hessian_approx = np.dot(gradient.T, gradient)

        else:
            criterion_candidate, gradient = fun_and_jac(x_candidate)

        # Line search
        if np.sum(criterion_candidate) > np.sum(criterion_accepted):
            step_size /= 2

            if step_size <= 0.01:
                # Accept step
                x_accepted = x_candidate
                criterion_accepted = criterion_candidate

                # Reset step size
                step_size = initial_step_size

        # If decrease in likelihood, calculate new direction vector
        else:
            # Accept step
            x_accepted = x_candidate
            criterion_accepted = criterion_candidate

            gradient_sum = np.sum(gradient, axis=0)
            direction = np.linalg.solve(hessian_approx, gradient_sum)
            gtol = np.dot(gradient_sum, direction)

            if gtol < 0:
                hessian_approx = np.dot(gradient.T, gradient)
                direction = np.linalg.solve(hessian_approx, gradient_sum)

            # Reset stepsize
            step_size = initial_step_size

        if gtol < gtol_abs:
            break

    res = InternalOptimizeResult(
        x=x_accepted,
        fun=criterion_accepted,
        message="Under development",
        n_iterations=niter,
    )

    return res



import numpy as np

CONVERGENCE_FTOL_REL = 2e-9
"""float: Stop when the relative improvement between two iterations is below this.

    The exact definition of relative improvement depends on the optimizer and should
    be documented there. To disable it, set it to 0.

    The default value is inspired by scipy L-BFGS-B defaults, but rounded.

"""

CONVERGENCE_FTOL_ABS = 0
"""float: Stop when the absolute improvement between two iterations is below this.

    Disabled by default because it is very problem specific.

"""

CONVERGENCE_GTOL_ABS = 1e-5
"""float: Stop when the gradient are smaller than this.

    For some algorithms this criterion refers to all entries, for others to some norm.

    For bound constrained optimizers this typically refers to a projected gradient.
    The exact definition should be documented for each optimizer.

    The default is the same as scipy. To disable it, set it to zero.

"""

CONVERGENCE_GTOL_REL = 1e-8
"""float: Stop when the gradient, divided by the absolute value of the criterion
    function is smaller than this. For some algorithms this criterion refers to
    all entries, for others to some norm.For bound constrained optimizers this
    typically refers to a projected gradient. The exact definition should be documented
    for each optimizer. To disable it, set it to zero.

"""

CONVERGENCE_GTOL_SCALED = 1e-8
"""float: Stop when all entries (or for some algorithms the norm) of the gradient,
    divided by the norm of the gradient at start parameters is smaller than this.
    For bound constrained optimizers this typically refers to a projected gradient.
    The exact definition should be documented for each optimizer.
    To disable it, set it to zero.

"""

CONVERGENCE_XTOL_REL = 1e-5
"""float: Stop when the relative change in parameters is smaller than this.
    The exact definition of relative change and whether this refers to the maximum
    change or the average change depends on the algorithm and should be documented
    there. To disable it, set it to zero. The default is the same as in scipy.

"""

CONVERGENCE_XTOL_ABS = 0
"""float: Stop when the absolute change in parameters between two iterations is smaller
    than this. Whether this refers to the maximum change or the average change depends
    on the algorithm and should be documented there.

    Disabled by default because it is very problem specific. To enable it, set it to a
    value larger than zero.

"""


STOPPING_MAXFUN = 1_000_000
"""int:
    If the maximum number of function evaluation is reached, the optimization stops
    but we do not count this as successful convergence. The function evaluations used
    to evaluate a numerical gradient do not count for this.

"""


STOPPING_MAXFUN_GLOBAL = 1_000
"""int:
    If the maximum number of function evaluation is reached, the optimization stops
    but we do not count this as successful convergence. The function evaluations used
    to evaluate a numerical gradient do not count for this. Set to a lower number than
    STOPPING_MAX_CRITERION_EVALUATIONS for global optimizers.

"""


STOPPING_MAXITER = 1_000_000
"""int:
    If the maximum number of iterations is reached, the
    optimization stops, but we do not count this as successful convergence.
    The difference to ``max_criterion_evaluations`` is that one iteration might
    need several criterion evaluations, for example in a line search or to determine
    if the trust region radius has to be shrunk.

"""

CONVERGENCE_SECOND_BEST_FTOL_ABS = 1e-08
"""float: absolute criterion tolerance optimagic requires if no other stopping
criterion apart from max iterations etc. is available
this is taken from scipy (SLSQP's value, smaller than Nelder-Mead).

"""

CONVERGENCE_SECOND_BEST_XTOL_ABS = 1e-08
"""float: The absolute parameter tolerance optimagic requires if no other stopping
criterion apart from max iterations etc. is available. This is taken from pybobyqa.

"""


MAX_LINE_SEARCH_STEPS = 20
"""int: Inspired by scipy L-BFGS-B."""

LIMITED_MEMORY_STORAGE_LENGTH = 10
"""int: Taken from scipy L-BFGS-B."""


CONSTRAINTS_ABSOLUTE_TOLERANCE = 1e-5
"""float: Allowed tolerance of the equality and inequality constraints for values to be
considered 'feasible'.

"""


def get_population_size(population_size, x, lower_bound=10):
    """Default population size for genetic algorithms."""
    if population_size is None:
        population_size = int(np.clip(10 * (len(x) + 1), lower_bound, np.inf))
    else:
        population_size = int(population_size)
    return population_size



from dataclasses import dataclass
from functools import partial

import numpy as np
from numpy.typing import NDArray
from pybaum import tree_just_flatten

from optimagic.optimization.history import History
from optimagic.parameters.tree_registry import get_registry
from optimagic.typing import Direction


@dataclass(frozen=True)
class HistoryArrays:
    fun: NDArray[np.float64]
    params: NDArray[np.float64]
    time: NDArray[np.float64]
    monotone_fun: NDArray[np.float64]
    is_accepted: NDArray[np.bool_]

    @property
    def criterion(self) -> NDArray[np.float64]:
        return self.fun

    @property
    def monotone_criterion(self) -> NDArray[np.float64]:
        return self.monotone_fun

    def __getitem__(self, key: str) -> NDArray[np.float64] | NDArray[np.bool_]:
        return getattr(self, key)


def get_history_arrays(history: History, direction: Direction) -> HistoryArrays:
    # ==================================================================================
    # Handle deprecations for now
    # ==================================================================================
    assert direction in [Direction.MINIMIZE, Direction.MAXIMIZE]

    if isinstance(history, dict):
        parhist = history["params"]
        funhist = history["criterion"]
        timehist = history["runtime"]

    else:
        parhist = history.params
        funhist = history.fun
        timehist = history.time

    # ==================================================================================
    # Filter out evaluations that do not have a `fun` value
    # ==================================================================================

    parhist = [p for p, f in zip(parhist, funhist, strict=False) if f is not None]
    timehist = [t for t, f in zip(timehist, funhist, strict=False) if f is not None]
    funhist = [f for f in funhist if f is not None]

    # ==================================================================================

    is_flat = (
        len(parhist) > 0 and isinstance(parhist[0], np.ndarray) and parhist[0].ndim == 1
    )
    if is_flat:
        to_internal = lambda x: x.tolist()
    else:
        registry = get_registry(extended=True)
        to_internal = partial(tree_just_flatten, registry=registry)

    critvals = np.array(funhist)

    params = np.array([to_internal(p) for p in parhist])

    runtimes = np.array(timehist)

    if direction == Direction.MINIMIZE:
        monotone = np.minimum.accumulate(critvals)
        is_accepted = critvals <= monotone
    elif direction == Direction.MAXIMIZE:
        monotone = np.maximum.accumulate(critvals)
        is_accepted = critvals >= monotone

    out = HistoryArrays(
        fun=critvals,
        params=params,
        time=runtimes,
        monotone_fun=monotone,
        is_accepted=is_accepted,
    )
    return out



from dataclasses import dataclass, replace
from typing import Any

import numpy as np

from optimagic.optimization.algorithm import InternalOptimizeResult
from optimagic.optimization.convergence_report import get_convergence_report
from optimagic.optimization.optimize_result import MultistartInfo, OptimizeResult
from optimagic.parameters.conversion import Converter
from optimagic.typing import AggregationLevel, Direction, PyTree
from optimagic.utilities import isscalar


@dataclass(frozen=True)
class ExtraResultFields:
    """Fields for OptimizeResult that are not part of InternalOptimizeResult."""

    start_fun: float
    start_params: PyTree
    algorithm: str
    direction: Direction
    n_free: int


def process_single_result(
    raw_res: InternalOptimizeResult,
    converter: Converter,
    solver_type: AggregationLevel,
    extra_fields: ExtraResultFields,
) -> OptimizeResult:
    """Process an internal optimizer result."""
    params = converter.params_from_internal(raw_res.x)
    if isscalar(raw_res.fun):
        fun = float(raw_res.fun)
    elif solver_type == AggregationLevel.LIKELIHOOD:
        fun = float(np.sum(raw_res.fun))
    elif solver_type == AggregationLevel.LEAST_SQUARES:
        fun = np.dot(raw_res.fun, raw_res.fun)

    if extra_fields.direction == Direction.MAXIMIZE:
        fun = -fun

    if raw_res.history is not None:
        conv_report = get_convergence_report(
            history=raw_res.history, direction=extra_fields.direction
        )
    else:
        conv_report = None

    out = OptimizeResult(
        params=params,
        fun=fun,
        start_fun=extra_fields.start_fun,
        start_params=extra_fields.start_params,
        algorithm=extra_fields.algorithm,
        direction=extra_fields.direction.value,
        n_free=extra_fields.n_free,
        message=raw_res.message,
        success=raw_res.success,
        n_fun_evals=raw_res.n_fun_evals,
        n_jac_evals=raw_res.n_jac_evals,
        n_hess_evals=raw_res.n_hess_evals,
        n_iterations=raw_res.n_iterations,
        status=raw_res.status,
        jac=raw_res.jac,
        hess=raw_res.hess,
        hess_inv=raw_res.hess_inv,
        max_constraint_violation=raw_res.max_constraint_violation,
        history=raw_res.history,
        algorithm_output=raw_res.info,
        convergence_report=conv_report,
    )
    return out


def process_multistart_result(
    raw_res: InternalOptimizeResult,
    converter: Converter,
    solver_type: AggregationLevel,
    extra_fields: ExtraResultFields,
) -> OptimizeResult:
    """Process results of internal optimizers.

    Args:
        res (dict): Results dictionary of an internal optimizer or multistart optimizer.

    """
    if raw_res.multistart_info is None:
        raise ValueError("Multistart info is missing.")

    if isinstance(raw_res, str):
        res = _dummy_result_from_traceback(raw_res, extra_fields)
    else:
        res = process_single_result(
            raw_res=raw_res,
            converter=converter,
            solver_type=solver_type,
            extra_fields=extra_fields,
        )

        info = _process_multistart_info(
            raw_res.multistart_info,
            converter=converter,
            solver_type=solver_type,
            extra_fields=extra_fields,
        )

        # ==============================================================================
        # create a convergence report for the multistart optimization; This is not
        # the same as the convergence report for the individual local optimizations.
        # ==============================================================================
        crit_hist = [opt.fun for opt in info.local_optima]
        params_hist = [opt.params for opt in info.local_optima]
        time_hist = [np.nan for opt in info.local_optima]
        hist = {"criterion": crit_hist, "params": params_hist, "runtime": time_hist}

        conv_report = get_convergence_report(
            history=hist,
            direction=extra_fields.direction,
        )

        res.convergence_report = conv_report

        res.algorithm = f"multistart_{res.algorithm}"
        res.n_iterations = _sum_or_none([opt.n_iterations for opt in info.local_optima])

        res.n_fun_evals = _sum_or_none([opt.n_fun_evals for opt in info.local_optima])
        res.n_jac_evals = _sum_or_none([opt.n_jac_evals for opt in info.local_optima])

        res.multistart_info = info
    return res


def _process_multistart_info(
    info: dict[str, Any],
    converter: Converter,
    solver_type: AggregationLevel,
    extra_fields: ExtraResultFields,
) -> MultistartInfo:
    starts = [converter.params_from_internal(x) for x in info["start_parameters"]]

    optima = []
    for res, start in zip(info["local_optima"], starts, strict=False):
        replacements = {
            "start_params": start,
            "start_fun": None,
        }

        processed = process_single_result(
            res,
            converter=converter,
            solver_type=solver_type,
            extra_fields=replace(extra_fields, **replacements),
        )
        optima.append(processed)

    sample = [converter.params_from_internal(x) for x in info["exploration_sample"]]

    if extra_fields.direction == Direction.MINIMIZE:
        exploration_res = info["exploration_results"]
    else:
        exploration_res = [-res for res in info["exploration_results"]]

    return MultistartInfo(
        start_parameters=starts,
        local_optima=optima,
        exploration_sample=sample,
        exploration_results=exploration_res,
    )


def _dummy_result_from_traceback(
    candidate: str, extra_fields: ExtraResultFields
) -> OptimizeResult:
    out = OptimizeResult(
        params=extra_fields.start_params,
        fun=extra_fields.start_fun,
        start_fun=extra_fields.start_fun,
        start_params=extra_fields.start_params,
        algorithm=extra_fields.algorithm,
        direction=extra_fields.direction.value,
        n_free=extra_fields.n_free,
        message=candidate,
    )
    return out


def _sum_or_none(summands: list[int | None | float]) -> int | None:
    if any(s is None for s in summands):
        out = None
    else:
        out = int(np.array(summands).sum())
    return out



"""This module contains various decorators.

There are two kinds of decorators defined in this module which consists of either two or
three nested functions. The former are decorators without and the latter with arguments.

For more information on decorators, see this `guide
`_ on https://realpython.com

which
provides a comprehensive overview.

.. _guide:

https://realpython.com/primer-on-python-decorators/

"""

import functools
import warnings

from optimagic.exceptions import get_traceback


def catch(
    func=None,
    *,
    exception=Exception,
    exclude=(KeyboardInterrupt, SystemExit),
    onerror=None,
    default=None,
    warn=True,
    reraise=False,
):
    """Catch and handle exceptions.

    This decorator can be used with and without additional arguments.

    Args:
        exception (Exception or tuple): One or several exceptions that
            are caught and handled. By default all Exceptions are
            caught and handled.
        exclude (Exception or tuple): One or several exceptionts that
            are not caught. By default those are KeyboardInterrupt and
            SystemExit.
        onerror (None or Callable): Callable that takes an Exception
            as only argument. This is called when an exception occurs.
        default: Value that is returned when as the output of func when
            an exception occurs. Can be one of the following:
            - a constant
            - "__traceback__", in this case a string with a traceback is returned.
            - callable with the same signature as func.
        warn (bool): If True, the exception is converted to a warning.
        reraise (bool): If True, the exception is raised after handling it.

    """

    def decorator_catch(func):
        @functools.wraps(func)
        def wrapper_catch(*args, **kwargs):
            try:
                res = func(*args, **kwargs)
            except exclude:
                raise
            except exception as e:
                if onerror is not None:
                    onerror(e)

                if reraise:
                    raise e

                tb = get_traceback()

                if warn:
                    msg = f"The following exception was caught:\n\n{tb}"
                    warnings.warn(msg)

                if default == "__traceback__":
                    res = tb
                elif callable(default):
                    res = default(*args, **kwargs)
                else:
                    res = default
            return res

        return wrapper_catch

    if callable(func):
        return decorator_catch(func)
    else:
        return decorator_catch


def unpack(func=None, symbol=None):
    def decorator_unpack(func):
        if symbol is None:

            @functools.wraps(func)
            def wrapper_unpack(arg):
                return func(arg)

        elif symbol == "*":

            @functools.wraps(func)
            def wrapper_unpack(arg):
                return func(*arg)

        elif symbol == "**":

            @functools.wraps(func)
            def wrapper_unpack(arg):
                return func(**arg)

        return wrapper_unpack

    if callable(func):
        return decorator_unpack(func)
    else:
        return decorator_unpack


def deprecated(func, msg):
    def decorator_deprecated(func):
        @functools.wraps(func)
        def wrapper_deprecated(*args, **kwargs):
            warnings.warn(msg, FutureWarning)
            return func(*args, **kwargs)

        return wrapper_deprecated

    if callable(func):
        return decorator_deprecated(func)
    else:
        return decorator_deprecated
